from datetime import date, datetime
from uuid import UUID

import sqlalchemy.orm as orm
from sqlalchemy.orm import Mapped

from gen_epix.commondb.repositories.sa_model import (
    NoIdRowMetadataMixin,
    create_mapped_column,
    create_table_args,
)
from gen_epix.omopdb.domain import DOMAIN, enum, model
from gen_epix.omopdb.repositories.sa_model.base import DataLineageMixin

Base: type = orm.declarative_base(name=enum.ServiceType.OMOP.value)


class Location(Base, NoIdRowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.Location)

    location_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Location, "location_id"
    )
    address_1: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Location, "address_1"
    )
    address_2: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Location, "address_2"
    )
    city: Mapped[str | None] = create_mapped_column(DOMAIN, model.Location, "city")
    state: Mapped[str | None] = create_mapped_column(DOMAIN, model.Location, "state")
    zip: Mapped[str | None] = create_mapped_column(DOMAIN, model.Location, "zip")
    county: Mapped[str | None] = create_mapped_column(DOMAIN, model.Location, "county")
    location_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Location, "location_source_value"
    )
    latitude: Mapped[float | None] = create_mapped_column(
        DOMAIN, model.Location, "latitude"
    )
    longitude: Mapped[float | None] = create_mapped_column(
        DOMAIN, model.Location, "longitude"
    )


class CohortDefinition(Base, NoIdRowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.CohortDefinition)

    cohort_definition_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.CohortDefinition, "cohort_definition_id"
    )
    cohort_definition_name: Mapped[str] = create_mapped_column(
        DOMAIN, model.CohortDefinition, "cohort_definition_name"
    )
    cohort_definition_description: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.CohortDefinition, "cohort_definition_description"
    )
    definition_type_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.CohortDefinition, "definition_type_concept_id"
    )
    cohort_definition_syntax: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.CohortDefinition, "cohort_definition_syntax"
    )
    subject_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.CohortDefinition, "subject_concept_id"
    )
    cohort_initiation_date: Mapped[date | None] = create_mapped_column(
        DOMAIN, model.CohortDefinition, "cohort_initiation_date"
    )


class Cohort(Base, NoIdRowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.Cohort)

    cohort_definition_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Cohort, "cohort_definition_id"
    )
    subject_id: Mapped[UUID] = create_mapped_column(DOMAIN, model.Cohort, "subject_id")
    cohort_start_date: Mapped[date] = create_mapped_column(
        DOMAIN, model.Cohort, "cohort_start_date"
    )
    cohort_end_date: Mapped[date] = create_mapped_column(
        DOMAIN, model.Cohort, "cohort_end_date"
    )
    cohort_id: Mapped[UUID] = create_mapped_column(DOMAIN, model.Cohort, "cohort_id")


class CdmSource(Base, NoIdRowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.CdmSource)

    cdm_source_name: Mapped[str] = create_mapped_column(
        DOMAIN, model.CdmSource, "cdm_source_name"
    )
    cdm_source_abbreviation: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.CdmSource, "cdm_source_abbreviation"
    )
    cdm_holder: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.CdmSource, "cdm_holder"
    )
    source_description: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.CdmSource, "source_description"
    )
    source_documentation_reference: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.CdmSource, "source_documentation_reference"
    )
    cdm_etl_reference: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.CdmSource, "cdm_etl_reference"
    )
    source_release_date: Mapped[date | None] = create_mapped_column(
        DOMAIN, model.CdmSource, "source_release_date"
    )
    cdm_release_date: Mapped[date | None] = create_mapped_column(
        DOMAIN, model.CdmSource, "cdm_release_date"
    )
    cdm_version: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.CdmSource, "cdm_version"
    )
    vocabulary_version: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.CdmSource, "vocabulary_version"
    )
    cdm_source_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.CdmSource, "cdm_source_id"
    )


class Vocabulary(Base, NoIdRowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.Vocabulary)

    vocabulary_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Vocabulary, "vocabulary_id"
    )
    vocabulary_name: Mapped[str] = create_mapped_column(
        DOMAIN, model.Vocabulary, "vocabulary_name"
    )
    vocabulary_reference: Mapped[str] = create_mapped_column(
        DOMAIN, model.Vocabulary, "vocabulary_reference"
    )
    vocabulary_version: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Vocabulary, "vocabulary_version"
    )
    vocabulary_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Vocabulary, "vocabulary_concept_id"
    )


class Domain(Base, NoIdRowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.Domain)

    domain_id: Mapped[UUID] = create_mapped_column(DOMAIN, model.Domain, "domain_id")
    domain_name: Mapped[str] = create_mapped_column(DOMAIN, model.Domain, "domain_name")
    domain_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Domain, "domain_concept_id"
    )


class ConceptClass(Base, NoIdRowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.ConceptClass)

    concept_class_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.ConceptClass, "concept_class_id"
    )
    concept_class_name: Mapped[str] = create_mapped_column(
        DOMAIN, model.ConceptClass, "concept_class_name"
    )
    concept_class_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.ConceptClass, "concept_class_concept_id"
    )


class Concept(Base, NoIdRowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.Concept)

    concept_id: Mapped[UUID] = create_mapped_column(DOMAIN, model.Concept, "concept_id")
    concept_name: Mapped[str] = create_mapped_column(
        DOMAIN, model.Concept, "concept_name"
    )
    domain_id: Mapped[UUID] = create_mapped_column(DOMAIN, model.Concept, "domain_id")
    vocabulary_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Concept, "vocabulary_id"
    )
    concept_class_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Concept, "concept_class_id"
    )
    standard_concept: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Concept, "standard_concept"
    )
    concept_code: Mapped[str] = create_mapped_column(
        DOMAIN, model.Concept, "concept_code"
    )
    valid_start_date: Mapped[date] = create_mapped_column(
        DOMAIN, model.Concept, "valid_start_date"
    )
    valid_end_date: Mapped[date] = create_mapped_column(
        DOMAIN, model.Concept, "valid_end_date"
    )
    invalid_reason: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Concept, "invalid_reason"
    )


class Relationship(Base, NoIdRowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.Relationship)

    relationship_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Relationship, "relationship_id"
    )
    relationship_name: Mapped[str] = create_mapped_column(
        DOMAIN, model.Relationship, "relationship_name"
    )
    is_hierarchical: Mapped[str] = create_mapped_column(
        DOMAIN, model.Relationship, "is_hierarchical"
    )
    defines_ancestry: Mapped[str] = create_mapped_column(
        DOMAIN, model.Relationship, "defines_ancestry"
    )
    reverse_relationship_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Relationship, "reverse_relationship_id"
    )
    relationship_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Relationship, "relationship_concept_id"
    )


class ConceptRelationship(Base, NoIdRowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.ConceptRelationship)

    concept_id_1: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.ConceptRelationship, "concept_id_1"
    )
    concept_id_2: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.ConceptRelationship, "concept_id_2"
    )
    relationship_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.ConceptRelationship, "relationship_id"
    )
    valid_start_date: Mapped[date] = create_mapped_column(
        DOMAIN, model.ConceptRelationship, "valid_start_date"
    )
    valid_end_date: Mapped[date] = create_mapped_column(
        DOMAIN, model.ConceptRelationship, "valid_end_date"
    )
    invalid_reason: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.ConceptRelationship, "invalid_reason"
    )
    concept_relationship_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.ConceptRelationship, "concept_relationship_id"
    )


class ConceptAncestor(Base, NoIdRowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.ConceptAncestor)

    ancestor_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.ConceptAncestor, "ancestor_concept_id"
    )
    descendant_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.ConceptAncestor, "descendant_concept_id"
    )
    min_levels_of_separation: Mapped[int] = create_mapped_column(
        DOMAIN, model.ConceptAncestor, "min_levels_of_separation"
    )
    max_levels_of_separation: Mapped[int] = create_mapped_column(
        DOMAIN, model.ConceptAncestor, "max_levels_of_separation"
    )
    concept_ancestor_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.ConceptAncestor, "concept_ancestor_id"
    )


class ConceptSynonym(Base, NoIdRowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.ConceptSynonym)

    concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.ConceptSynonym, "concept_id"
    )
    concept_synonym_name: Mapped[str] = create_mapped_column(
        DOMAIN, model.ConceptSynonym, "concept_synonym_name"
    )
    language_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.ConceptSynonym, "language_concept_id"
    )
    concept_synonym_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.ConceptSynonym, "concept_synonym_id"
    )


class DrugStrength(Base, NoIdRowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.DrugStrength)

    drug_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.DrugStrength, "drug_concept_id"
    )
    ingredient_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.DrugStrength, "ingredient_concept_id"
    )
    amount_value: Mapped[float | None] = create_mapped_column(
        DOMAIN, model.DrugStrength, "amount_value"
    )
    amount_unit_concept_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.DrugStrength, "amount_unit_concept_id"
    )
    numerator_value: Mapped[float | None] = create_mapped_column(
        DOMAIN, model.DrugStrength, "numerator_value"
    )
    numerator_unit_concept_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.DrugStrength, "numerator_unit_concept_id"
    )
    denominator_value: Mapped[float | None] = create_mapped_column(
        DOMAIN, model.DrugStrength, "denominator_value"
    )
    denominator_unit_concept_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.DrugStrength, "denominator_unit_concept_id"
    )
    box_size: Mapped[int | None] = create_mapped_column(
        DOMAIN, model.DrugStrength, "box_size"
    )
    valid_start_date: Mapped[date] = create_mapped_column(
        DOMAIN, model.DrugStrength, "valid_start_date"
    )
    valid_end_date: Mapped[date] = create_mapped_column(
        DOMAIN, model.DrugStrength, "valid_end_date"
    )
    invalid_reason: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.DrugStrength, "invalid_reason"
    )
    drug_strength_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.DrugStrength, "drug_strength_id"
    )


class SourceToConceptMap(Base, NoIdRowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.SourceToConceptMap)

    source_code: Mapped[str] = create_mapped_column(
        DOMAIN, model.SourceToConceptMap, "source_code"
    )
    source_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.SourceToConceptMap, "source_concept_id"
    )
    source_vocabulary_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.SourceToConceptMap, "source_vocabulary_id"
    )
    source_code_description: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.SourceToConceptMap, "source_code_description"
    )
    target_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.SourceToConceptMap, "target_concept_id"
    )
    target_vocabulary_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.SourceToConceptMap, "target_vocabulary_id"
    )
    valid_start_date: Mapped[date] = create_mapped_column(
        DOMAIN, model.SourceToConceptMap, "valid_start_date"
    )
    valid_end_date: Mapped[date] = create_mapped_column(
        DOMAIN, model.SourceToConceptMap, "valid_end_date"
    )
    invalid_reason: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.SourceToConceptMap, "invalid_reason"
    )
    source_to_concept_map_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.SourceToConceptMap, "source_to_concept_map_id"
    )


class Metadata(Base, NoIdRowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.Metadata)

    metadata_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Metadata, "metadata_concept_id"
    )
    metadata_type_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Metadata, "metadata_type_concept_id"
    )
    name: Mapped[str] = create_mapped_column(DOMAIN, model.Metadata, "name")
    value_as_string: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Metadata, "value_as_string"
    )
    value_as_concept_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Metadata, "value_as_concept_id"
    )
    metadata_date: Mapped[date | None] = create_mapped_column(
        DOMAIN, model.Metadata, "metadata_date"
    )
    metadata_datetime: Mapped[datetime | None] = create_mapped_column(
        DOMAIN, model.Metadata, "metadata_datetime"
    )
    metadata_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Metadata, "metadata_id"
    )


class CareSite(Base, NoIdRowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.CareSite)

    care_site_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.CareSite, "care_site_id"
    )
    care_site_name: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.CareSite, "care_site_name"
    )
    place_of_service_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.CareSite, "place_of_service_concept_id"
    )
    location_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.CareSite, "location_id"
    )
    care_site_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.CareSite, "care_site_source_value"
    )
    place_of_service_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.CareSite, "place_of_service_source_value"
    )
    site_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.CareSite, "site_id"
    )


class Provider(Base, NoIdRowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.Provider)

    provider_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Provider, "provider_id"
    )
    provider_name: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Provider, "provider_name"
    )
    npi: Mapped[str | None] = create_mapped_column(DOMAIN, model.Provider, "npi")
    dea: Mapped[str | None] = create_mapped_column(DOMAIN, model.Provider, "dea")
    specialty_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Provider, "specialty_concept_id"
    )
    care_site_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Provider, "care_site_id"
    )
    year_of_birth: Mapped[int | None] = create_mapped_column(
        DOMAIN, model.Provider, "year_of_birth"
    )
    gender_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Provider, "gender_concept_id"
    )
    provider_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Provider, "provider_source_value"
    )
    specialty_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Provider, "specialty_source_value"
    )
    specialty_source_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Provider, "specialty_source_concept_id"
    )
    gender_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Provider, "gender_source_value"
    )
    gender_source_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Provider, "gender_source_concept_id"
    )


class Person(Base, DataLineageMixin, NoIdRowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.Person)

    person_id: Mapped[UUID] = create_mapped_column(DOMAIN, model.Person, "person_id")
    gender_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Person, "gender_concept_id"
    )
    year_of_birth: Mapped[int] = create_mapped_column(
        DOMAIN, model.Person, "year_of_birth"
    )
    month_of_birth: Mapped[int | None] = create_mapped_column(
        DOMAIN, model.Person, "month_of_birth"
    )
    day_of_birth: Mapped[int | None] = create_mapped_column(
        DOMAIN, model.Person, "day_of_birth"
    )
    birth_datetime: Mapped[datetime | None] = create_mapped_column(
        DOMAIN, model.Person, "birth_datetime"
    )
    death_datetime: Mapped[datetime | None] = create_mapped_column(
        DOMAIN, model.Person, "death_datetime"
    )
    race_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Person, "race_concept_id"
    )
    ethnicity_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Person, "ethnicity_concept_id"
    )
    location_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Person, "location_id"
    )
    provider_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Person, "provider_id"
    )
    care_site_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Person, "care_site_id"
    )
    person_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Person, "person_source_value"
    )
    gender_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Person, "gender_source_value"
    )
    gender_source_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Person, "gender_source_concept_id"
    )
    race_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Person, "race_source_value"
    )
    race_source_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Person, "race_source_concept_id"
    )
    ethnicity_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Person, "ethnicity_source_value"
    )
    ethnicity_source_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Person, "ethnicity_source_concept_id"
    )
    provided_by_organization_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Person, "provided_by_organization_id"
    )
    person_type_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Person, "person_type_concept_id"
    )


class ObservationPeriod(Base, DataLineageMixin, NoIdRowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.ObservationPeriod)

    observation_period_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.ObservationPeriod, "observation_period_id"
    )
    person_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.ObservationPeriod, "person_id"
    )
    observation_period_start_date: Mapped[date] = create_mapped_column(
        DOMAIN, model.ObservationPeriod, "observation_period_start_date"
    )
    observation_period_end_date: Mapped[date] = create_mapped_column(
        DOMAIN, model.ObservationPeriod, "observation_period_end_date"
    )
    period_type_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.ObservationPeriod, "period_type_concept_id"
    )
    observation_period_start_iso_interval: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.ObservationPeriod, "observation_period_start_iso_interval"
    )
    observation_period_end_iso_interval: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.ObservationPeriod, "observation_period_end_iso_interval"
    )


class PayerPlanPeriod(Base, DataLineageMixin, NoIdRowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.PayerPlanPeriod)

    payer_plan_period_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.PayerPlanPeriod, "payer_plan_period_id"
    )
    person_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.PayerPlanPeriod, "person_id"
    )
    contract_person_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.PayerPlanPeriod, "contract_person_id"
    )
    payer_plan_period_start_date: Mapped[date] = create_mapped_column(
        DOMAIN, model.PayerPlanPeriod, "payer_plan_period_start_date"
    )
    payer_plan_period_end_date: Mapped[date] = create_mapped_column(
        DOMAIN, model.PayerPlanPeriod, "payer_plan_period_end_date"
    )
    payer_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.PayerPlanPeriod, "payer_concept_id"
    )
    payer_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.PayerPlanPeriod, "payer_source_value"
    )
    payer_source_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.PayerPlanPeriod, "payer_source_concept_id"
    )
    plan_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.PayerPlanPeriod, "plan_concept_id"
    )
    plan_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.PayerPlanPeriod, "plan_source_value"
    )
    plan_source_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.PayerPlanPeriod, "plan_source_concept_id"
    )
    contract_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.PayerPlanPeriod, "contract_concept_id"
    )
    contract_source_value: Mapped[str] = create_mapped_column(
        DOMAIN, model.PayerPlanPeriod, "contract_source_value"
    )
    contract_source_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.PayerPlanPeriod, "contract_source_concept_id"
    )
    sponsor_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.PayerPlanPeriod, "sponsor_concept_id"
    )
    sponsor_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.PayerPlanPeriod, "sponsor_source_value"
    )
    sponsor_source_concept_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.PayerPlanPeriod, "sponsor_source_concept_id"
    )
    family_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.PayerPlanPeriod, "family_source_value"
    )
    stop_reason_concept_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.PayerPlanPeriod, "stop_reason_concept_id"
    )
    stop_reason_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.PayerPlanPeriod, "stop_reason_source_value"
    )
    stop_reason_source_concept_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.PayerPlanPeriod, "stop_reason_source_concept_id"
    )


class VisitOccurrence(Base, DataLineageMixin, NoIdRowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.VisitOccurrence)

    visit_occurrence_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.VisitOccurrence, "visit_occurrence_id"
    )
    person_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.VisitOccurrence, "person_id"
    )
    visit_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.VisitOccurrence, "visit_concept_id"
    )
    visit_start_date: Mapped[date | None] = create_mapped_column(
        DOMAIN, model.VisitOccurrence, "visit_start_date"
    )
    visit_start_datetime: Mapped[datetime] = create_mapped_column(
        DOMAIN, model.VisitOccurrence, "visit_start_datetime"
    )
    visit_end_date: Mapped[date | None] = create_mapped_column(
        DOMAIN, model.VisitOccurrence, "visit_end_date"
    )
    visit_end_datetime: Mapped[datetime] = create_mapped_column(
        DOMAIN, model.VisitOccurrence, "visit_end_datetime"
    )
    visit_type_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.VisitOccurrence, "visit_type_concept_id"
    )
    provider_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.VisitOccurrence, "provider_id"
    )
    care_site_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.VisitOccurrence, "care_site_id"
    )
    visit_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.VisitOccurrence, "visit_source_value"
    )
    visit_source_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.VisitOccurrence, "visit_source_concept_id"
    )
    admitted_from_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.VisitOccurrence, "admitted_from_concept_id"
    )
    admitted_from_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.VisitOccurrence, "admitted_from_source_value"
    )
    discharge_to_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.VisitOccurrence, "discharge_to_concept_id"
    )
    discharge_to_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.VisitOccurrence, "discharge_to_source_value"
    )
    preceding_visit_occurrence_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.VisitOccurrence, "preceding_visit_occurrence_id"
    )


class VisitDetail(Base, DataLineageMixin, NoIdRowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.VisitDetail)

    visit_detail_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.VisitDetail, "visit_detail_id"
    )
    person_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.VisitDetail, "person_id"
    )
    visit_detail_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.VisitDetail, "visit_detail_concept_id"
    )
    visit_detail_start_date: Mapped[date] = create_mapped_column(
        DOMAIN, model.VisitDetail, "visit_detail_start_date"
    )
    visit_detail_start_datetime: Mapped[datetime | None] = create_mapped_column(
        DOMAIN, model.VisitDetail, "visit_detail_start_datetime"
    )
    visit_detail_end_date: Mapped[date] = create_mapped_column(
        DOMAIN, model.VisitDetail, "visit_detail_end_date"
    )
    visit_detail_end_datetime: Mapped[datetime | None] = create_mapped_column(
        DOMAIN, model.VisitDetail, "visit_detail_end_datetime"
    )
    visit_detail_type_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.VisitDetail, "visit_detail_type_concept_id"
    )
    provider_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.VisitDetail, "provider_id"
    )
    care_site_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.VisitDetail, "care_site_id"
    )
    visit_detail_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.VisitDetail, "visit_detail_source_value"
    )
    visit_detail_source_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.VisitDetail, "visit_detail_source_concept_id"
    )
    admitted_from_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.VisitDetail, "admitted_from_source_value"
    )
    admitted_from_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.VisitDetail, "admitted_from_concept_id"
    )
    discharge_to_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.VisitDetail, "discharge_to_source_value"
    )
    discharge_to_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.VisitDetail, "discharge_to_concept_id"
    )
    preceding_visit_detail_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.VisitDetail, "preceding_visit_detail_id"
    )
    visit_detail_parent_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.VisitDetail, "visit_detail_parent_id"
    )
    visit_occurrence_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.VisitDetail, "visit_occurrence_id"
    )


class ConditionOccurrence(Base, DataLineageMixin, NoIdRowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.ConditionOccurrence)

    condition_occurrence_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.ConditionOccurrence, "condition_occurrence_id"
    )
    person_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.ConditionOccurrence, "person_id"
    )
    condition_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.ConditionOccurrence, "condition_concept_id"
    )
    condition_start_date: Mapped[date] = create_mapped_column(
        DOMAIN, model.ConditionOccurrence, "condition_start_date"
    )
    condition_start_datetime: Mapped[datetime | None] = create_mapped_column(
        DOMAIN, model.ConditionOccurrence, "condition_start_datetime"
    )
    condition_end_date: Mapped[date | None] = create_mapped_column(
        DOMAIN, model.ConditionOccurrence, "condition_end_date"
    )
    condition_end_datetime: Mapped[datetime | None] = create_mapped_column(
        DOMAIN, model.ConditionOccurrence, "condition_end_datetime"
    )
    condition_type_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.ConditionOccurrence, "condition_type_concept_id"
    )
    condition_status_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.ConditionOccurrence, "condition_status_concept_id"
    )
    stop_reason: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.ConditionOccurrence, "stop_reason"
    )
    provider_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.ConditionOccurrence, "provider_id"
    )
    visit_occurrence_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.ConditionOccurrence, "visit_occurrence_id"
    )
    visit_detail_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.ConditionOccurrence, "visit_detail_id"
    )
    condition_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.ConditionOccurrence, "condition_source_value"
    )
    condition_source_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.ConditionOccurrence, "condition_source_concept_id"
    )
    condition_status_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.ConditionOccurrence, "condition_status_source_value"
    )
    condition_start_iso_interval: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.ConditionOccurrence, "condition_start_iso_interval"
    )
    condition_end_iso_interval: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.ConditionOccurrence, "condition_end_iso_interval"
    )


class ProcedureOccurrence(Base, DataLineageMixin, NoIdRowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.ProcedureOccurrence)

    procedure_occurrence_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.ProcedureOccurrence, "procedure_occurrence_id"
    )
    person_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.ProcedureOccurrence, "person_id"
    )
    procedure_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.ProcedureOccurrence, "procedure_concept_id"
    )
    procedure_date: Mapped[date | None] = create_mapped_column(
        DOMAIN, model.ProcedureOccurrence, "procedure_date"
    )
    procedure_datetime: Mapped[datetime] = create_mapped_column(
        DOMAIN, model.ProcedureOccurrence, "procedure_datetime"
    )
    procedure_type_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.ProcedureOccurrence, "procedure_type_concept_id"
    )
    modifier_concept_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.ProcedureOccurrence, "modifier_concept_id"
    )
    quantity: Mapped[int | None] = create_mapped_column(
        DOMAIN, model.ProcedureOccurrence, "quantity"
    )
    provider_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.ProcedureOccurrence, "provider_id"
    )
    visit_occurrence_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.ProcedureOccurrence, "visit_occurrence_id"
    )
    visit_detail_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.ProcedureOccurrence, "visit_detail_id"
    )
    procedure_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.ProcedureOccurrence, "procedure_source_value"
    )
    procedure_source_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.ProcedureOccurrence, "procedure_source_concept_id"
    )
    modifier_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.ProcedureOccurrence, "modifier_source_value"
    )
    procedure_iso_interval: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.ProcedureOccurrence, "procedure_iso_interval"
    )


class DrugExposure(Base, DataLineageMixin, NoIdRowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.DrugExposure)

    drug_exposure_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.DrugExposure, "drug_exposure_id"
    )
    person_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.DrugExposure, "person_id"
    )
    drug_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.DrugExposure, "drug_concept_id"
    )
    drug_exposure_start_date: Mapped[date] = create_mapped_column(
        DOMAIN, model.DrugExposure, "drug_exposure_start_date"
    )
    drug_exposure_start_datetime: Mapped[datetime | None] = create_mapped_column(
        DOMAIN, model.DrugExposure, "drug_exposure_start_datetime"
    )
    drug_exposure_end_date: Mapped[date] = create_mapped_column(
        DOMAIN, model.DrugExposure, "drug_exposure_end_date"
    )
    drug_exposure_end_datetime: Mapped[datetime | None] = create_mapped_column(
        DOMAIN, model.DrugExposure, "drug_exposure_end_datetime"
    )
    verbatim_end_date: Mapped[date | None] = create_mapped_column(
        DOMAIN, model.DrugExposure, "verbatim_end_date"
    )
    drug_type_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.DrugExposure, "drug_type_concept_id"
    )
    stop_reason: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.DrugExposure, "stop_reason"
    )
    refills: Mapped[int | None] = create_mapped_column(
        DOMAIN, model.DrugExposure, "refills"
    )
    quantity: Mapped[float | None] = create_mapped_column(
        DOMAIN, model.DrugExposure, "quantity"
    )
    days_supply: Mapped[int | None] = create_mapped_column(
        DOMAIN, model.DrugExposure, "days_supply"
    )
    sig: Mapped[str | None] = create_mapped_column(DOMAIN, model.DrugExposure, "sig")
    route_concept_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.DrugExposure, "route_concept_id"
    )
    lot_number: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.DrugExposure, "lot_number"
    )
    provider_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.DrugExposure, "provider_id"
    )
    visit_occurrence_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.DrugExposure, "visit_occurrence_id"
    )
    visit_detail_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.DrugExposure, "visit_detail_id"
    )
    drug_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.DrugExposure, "drug_source_value"
    )
    drug_source_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.DrugExposure, "drug_source_concept_id"
    )
    route_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.DrugExposure, "route_source_value"
    )
    dose_unit_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.DrugExposure, "dose_unit_source_value"
    )
    drug_exposure_start_iso_interval: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.DrugExposure, "drug_exposure_start_iso_interval"
    )
    drug_exposure_end_iso_interval: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.DrugExposure, "drug_exposure_end_iso_interval"
    )


class DeviceExposure(Base, DataLineageMixin, NoIdRowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.DeviceExposure)

    device_exposure_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.DeviceExposure, "device_exposure_id"
    )
    person_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.DeviceExposure, "person_id"
    )
    device_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.DeviceExposure, "device_concept_id"
    )
    device_exposure_start_date: Mapped[date] = create_mapped_column(
        DOMAIN, model.DeviceExposure, "device_exposure_start_date"
    )
    device_exposure_start_datetime: Mapped[datetime | None] = create_mapped_column(
        DOMAIN, model.DeviceExposure, "device_exposure_start_datetime"
    )
    device_exposure_end_date: Mapped[date | None] = create_mapped_column(
        DOMAIN, model.DeviceExposure, "device_exposure_end_date"
    )
    device_exposure_end_datetime: Mapped[datetime | None] = create_mapped_column(
        DOMAIN, model.DeviceExposure, "device_exposure_end_datetime"
    )
    device_type_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.DeviceExposure, "device_type_concept_id"
    )
    unique_device_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.DeviceExposure, "unique_device_id"
    )
    quantity: Mapped[int | None] = create_mapped_column(
        DOMAIN, model.DeviceExposure, "quantity"
    )
    provider_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.DeviceExposure, "provider_id"
    )
    visit_occurrence_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.DeviceExposure, "visit_occurrence_id"
    )
    visit_detail_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.DeviceExposure, "visit_detail_id"
    )
    device_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.DeviceExposure, "device_source_value"
    )
    device_source_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.DeviceExposure, "device_source_concept_id"
    )
    device_exposure_start_iso_interval: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.DeviceExposure, "device_exposure_start_iso_interval"
    )
    device_exposure_end_iso_interval: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.DeviceExposure, "device_exposure_end_iso_interval"
    )


class Measurement(Base, DataLineageMixin, NoIdRowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.Measurement)

    measurement_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Measurement, "measurement_id"
    )
    person_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Measurement, "person_id"
    )
    measurement_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Measurement, "measurement_concept_id"
    )
    measurement_date: Mapped[date] = create_mapped_column(
        DOMAIN, model.Measurement, "measurement_date"
    )
    measurement_datetime: Mapped[datetime | None] = create_mapped_column(
        DOMAIN, model.Measurement, "measurement_datetime"
    )
    measurement_time: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Measurement, "measurement_time"
    )
    measurement_type_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Measurement, "measurement_type_concept_id"
    )
    operator_concept_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Measurement, "operator_concept_id"
    )
    value_as_number: Mapped[float | None] = create_mapped_column(
        DOMAIN, model.Measurement, "value_as_number"
    )
    value_as_concept_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Measurement, "value_as_concept_id"
    )
    unit_concept_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Measurement, "unit_concept_id"
    )
    range_low: Mapped[float | None] = create_mapped_column(
        DOMAIN, model.Measurement, "range_low"
    )
    range_high: Mapped[float | None] = create_mapped_column(
        DOMAIN, model.Measurement, "range_high"
    )
    provider_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Measurement, "provider_id"
    )
    visit_occurrence_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Measurement, "visit_occurrence_id"
    )
    visit_detail_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Measurement, "visit_detail_id"
    )
    measurement_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Measurement, "measurement_source_value"
    )
    measurement_source_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Measurement, "measurement_source_concept_id"
    )
    unit_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Measurement, "unit_source_value"
    )
    value_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Measurement, "value_source_value"
    )
    measurement_iso_interval: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Measurement, "measurement_iso_interval"
    )
    derived_from_specimen_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Measurement, "derived_from_specimen_id"
    )


class Observation(Base, DataLineageMixin, NoIdRowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.Observation)

    observation_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Observation, "observation_id"
    )
    person_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Observation, "person_id"
    )
    observation_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Observation, "observation_concept_id"
    )
    observation_date: Mapped[date | None] = create_mapped_column(
        DOMAIN, model.Observation, "observation_date"
    )
    observation_datetime: Mapped[datetime] = create_mapped_column(
        DOMAIN, model.Observation, "observation_datetime"
    )
    observation_type_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Observation, "observation_type_concept_id"
    )
    value_as_number: Mapped[float | None] = create_mapped_column(
        DOMAIN, model.Observation, "value_as_number"
    )
    value_as_string: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Observation, "value_as_string"
    )
    value_as_concept_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Observation, "value_as_concept_id"
    )
    qualifier_concept_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Observation, "qualifier_concept_id"
    )
    unit_concept_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Observation, "unit_concept_id"
    )
    provider_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Observation, "provider_id"
    )
    visit_occurrence_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Observation, "visit_occurrence_id"
    )
    visit_detail_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Observation, "visit_detail_id"
    )
    observation_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Observation, "observation_source_value"
    )
    observation_source_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Observation, "observation_source_concept_id"
    )
    unit_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Observation, "unit_source_value"
    )
    qualifier_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Observation, "qualifier_source_value"
    )
    observation_event_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Observation, "observation_event_id"
    )
    obs_event_field_concept_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Observation, "obs_event_field_concept_id"
    )
    value_as_datetime: Mapped[datetime | None] = create_mapped_column(
        DOMAIN, model.Observation, "value_as_datetime"
    )
    observation_iso_interval: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Observation, "observation_iso_interval"
    )
    value_as_iso_interval: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Observation, "value_as_iso_interval"
    )


class Specimen(Base, DataLineageMixin, NoIdRowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.Specimen)

    specimen_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Specimen, "specimen_id"
    )
    person_id: Mapped[UUID] = create_mapped_column(DOMAIN, model.Specimen, "person_id")
    specimen_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Specimen, "specimen_concept_id"
    )
    specimen_type_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Specimen, "specimen_type_concept_id"
    )
    specimen_date: Mapped[date] = create_mapped_column(
        DOMAIN, model.Specimen, "specimen_date"
    )
    specimen_datetime: Mapped[datetime | None] = create_mapped_column(
        DOMAIN, model.Specimen, "specimen_datetime"
    )
    quantity: Mapped[float | None] = create_mapped_column(
        DOMAIN, model.Specimen, "quantity"
    )
    unit_concept_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Specimen, "unit_concept_id"
    )
    anatomic_site_concept_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Specimen, "anatomic_site_concept_id"
    )
    disease_status_concept_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Specimen, "disease_status_concept_id"
    )
    specimen_source_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Specimen, "specimen_source_id"
    )
    specimen_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Specimen, "specimen_source_value"
    )
    unit_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Specimen, "unit_source_value"
    )
    anatomic_site_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Specimen, "anatomic_site_source_value"
    )
    disease_status_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Specimen, "disease_status_source_value"
    )
    specimen_iso_interval: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Specimen, "specimen_iso_interval"
    )
    derived_from_specimen_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Specimen, "derived_from_specimen_id"
    )
    derived_from_specimen_concept_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Specimen, "derived_from_specimen_concept_id"
    )


class Note(Base, DataLineageMixin, NoIdRowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.Note)

    note_id: Mapped[UUID] = create_mapped_column(DOMAIN, model.Note, "note_id")
    person_id: Mapped[UUID] = create_mapped_column(DOMAIN, model.Note, "person_id")
    note_event_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Note, "note_event_id"
    )
    note_event_field_concept_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Note, "note_event_field_concept_id"
    )
    note_date: Mapped[date] = create_mapped_column(DOMAIN, model.Note, "note_date")
    note_datetime: Mapped[datetime | None] = create_mapped_column(
        DOMAIN, model.Note, "note_datetime"
    )
    note_type_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Note, "note_type_concept_id"
    )
    note_class_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Note, "note_class_concept_id"
    )
    note_title: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Note, "note_title"
    )
    note_text: Mapped[str] = create_mapped_column(DOMAIN, model.Note, "note_text")
    encoding_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Note, "encoding_concept_id"
    )
    language_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Note, "language_concept_id"
    )
    provider_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Note, "provider_id"
    )
    visit_occurrence_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Note, "visit_occurrence_id"
    )
    visit_detail_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Note, "visit_detail_id"
    )
    note_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Note, "note_source_value"
    )


class ConditionEra(Base, DataLineageMixin, NoIdRowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.ConditionEra)

    condition_era_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.ConditionEra, "condition_era_id"
    )
    person_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.ConditionEra, "person_id"
    )
    condition_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.ConditionEra, "condition_concept_id"
    )
    condition_era_start_datetime: Mapped[datetime] = create_mapped_column(
        DOMAIN, model.ConditionEra, "condition_era_start_datetime"
    )
    condition_era_end_datetime: Mapped[datetime] = create_mapped_column(
        DOMAIN, model.ConditionEra, "condition_era_end_datetime"
    )
    condition_occurrence_count: Mapped[int | None] = create_mapped_column(
        DOMAIN, model.ConditionEra, "condition_occurrence_count"
    )


class DrugEra(Base, DataLineageMixin, NoIdRowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.DrugEra)

    drug_era_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.DrugEra, "drug_era_id"
    )
    person_id: Mapped[UUID] = create_mapped_column(DOMAIN, model.DrugEra, "person_id")
    drug_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.DrugEra, "drug_concept_id"
    )
    drug_era_start_datetime: Mapped[datetime] = create_mapped_column(
        DOMAIN, model.DrugEra, "drug_era_start_datetime"
    )
    drug_era_end_datetime: Mapped[datetime] = create_mapped_column(
        DOMAIN, model.DrugEra, "drug_era_end_datetime"
    )
    drug_exposure_count: Mapped[int | None] = create_mapped_column(
        DOMAIN, model.DrugEra, "drug_exposure_count"
    )
    gap_days: Mapped[int | None] = create_mapped_column(
        DOMAIN, model.DrugEra, "gap_days"
    )
    drug_era_start_iso_interval: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.DrugEra, "drug_era_start_iso_interval"
    )
    drug_era_end_iso_interval: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.DrugEra, "drug_era_end_iso_interval"
    )


class DoseEra(Base, DataLineageMixin, NoIdRowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.DoseEra)

    dose_era_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.DoseEra, "dose_era_id"
    )
    person_id: Mapped[UUID] = create_mapped_column(DOMAIN, model.DoseEra, "person_id")
    drug_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.DoseEra, "drug_concept_id"
    )
    unit_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.DoseEra, "unit_concept_id"
    )
    dose_value: Mapped[float] = create_mapped_column(
        DOMAIN, model.DoseEra, "dose_value"
    )
    dose_era_start_datetime: Mapped[datetime] = create_mapped_column(
        DOMAIN, model.DoseEra, "dose_era_start_datetime"
    )
    dose_era_end_datetime: Mapped[datetime] = create_mapped_column(
        DOMAIN, model.DoseEra, "dose_era_end_datetime"
    )


class NoteNlp(Base, DataLineageMixin, NoIdRowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.NoteNlp)

    note_nlp_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.NoteNlp, "note_nlp_id"
    )
    note_id: Mapped[UUID] = create_mapped_column(DOMAIN, model.NoteNlp, "note_id")
    section_concept_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.NoteNlp, "section_concept_id"
    )
    snippet: Mapped[str | None] = create_mapped_column(DOMAIN, model.NoteNlp, "snippet")
    offset: Mapped[str | None] = create_mapped_column(DOMAIN, model.NoteNlp, "offset")
    lexical_variant: Mapped[str] = create_mapped_column(
        DOMAIN, model.NoteNlp, "lexical_variant"
    )
    note_nlp_concept_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.NoteNlp, "note_nlp_concept_id"
    )
    note_nlp_source_concept_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.NoteNlp, "note_nlp_source_concept_id"
    )
    nlp_system: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.NoteNlp, "nlp_system"
    )
    nlp_date: Mapped[date] = create_mapped_column(DOMAIN, model.NoteNlp, "nlp_date")
    nlp_datetime: Mapped[datetime | None] = create_mapped_column(
        DOMAIN, model.NoteNlp, "nlp_datetime"
    )
    term_exists: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.NoteNlp, "term_exists"
    )
    term_temporal: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.NoteNlp, "term_temporal"
    )
    term_modifiers: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.NoteNlp, "term_modifiers"
    )


class Cost(Base, DataLineageMixin, NoIdRowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.Cost)

    cost_id: Mapped[UUID] = create_mapped_column(DOMAIN, model.Cost, "cost_id")
    person_id: Mapped[UUID] = create_mapped_column(DOMAIN, model.Cost, "person_id")
    cost_event_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Cost, "cost_event_id"
    )
    cost_event_field_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Cost, "cost_event_field_concept_id"
    )
    cost_concept_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Cost, "cost_concept_id"
    )
    cost_type_concept_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Cost, "cost_type_concept_id"
    )
    cost_source_concept_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Cost, "cost_source_concept_id"
    )
    cost_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Cost, "cost_source_value"
    )
    currency_concept_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Cost, "currency_concept_id"
    )
    cost: Mapped[float | None] = create_mapped_column(DOMAIN, model.Cost, "cost")
    incurred_date: Mapped[date | None] = create_mapped_column(
        DOMAIN, model.Cost, "incurred_date"
    )
    billed_date: Mapped[date | None] = create_mapped_column(
        DOMAIN, model.Cost, "billed_date"
    )
    paid_date: Mapped[date | None] = create_mapped_column(
        DOMAIN, model.Cost, "paid_date"
    )
    revenue_code_concept_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Cost, "revenue_code_concept_id"
    )
    drg_concept_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Cost, "drg_concept_id"
    )
    revenue_code_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Cost, "revenue_code_source_value"
    )
    drg_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Cost, "drg_source_value"
    )
    payer_plan_period_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Cost, "payer_plan_period_id"
    )


class LocationHistory(Base, DataLineageMixin, NoIdRowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.LocationHistory)

    location_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.LocationHistory, "location_id"
    )
    relationship_type_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.LocationHistory, "relationship_type_concept_id"
    )
    domain_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.LocationHistory, "domain_id"
    )
    entity_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.LocationHistory, "entity_id"
    )
    start_date: Mapped[date] = create_mapped_column(
        DOMAIN, model.LocationHistory, "start_date"
    )
    end_date: Mapped[date | None] = create_mapped_column(
        DOMAIN, model.LocationHistory, "end_date"
    )
    start_iso_interval: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.LocationHistory, "start_iso_interval"
    )
    end_iso_interval: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.LocationHistory, "end_iso_interval"
    )
    location_history_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.LocationHistory, "location_history_id"
    )


class SurveyConduct(Base, DataLineageMixin, NoIdRowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.SurveyConduct)

    survey_conduct_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.SurveyConduct, "survey_conduct_id"
    )
    person_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.SurveyConduct, "person_id"
    )
    survey_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.SurveyConduct, "survey_concept_id"
    )
    survey_start_date: Mapped[date | None] = create_mapped_column(
        DOMAIN, model.SurveyConduct, "survey_start_date"
    )
    survey_start_datetime: Mapped[datetime | None] = create_mapped_column(
        DOMAIN, model.SurveyConduct, "survey_start_datetime"
    )
    survey_end_date: Mapped[date | None] = create_mapped_column(
        DOMAIN, model.SurveyConduct, "survey_end_date"
    )
    survey_end_datetime: Mapped[datetime] = create_mapped_column(
        DOMAIN, model.SurveyConduct, "survey_end_datetime"
    )
    provider_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.SurveyConduct, "provider_id"
    )
    assisted_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.SurveyConduct, "assisted_concept_id"
    )
    respondent_type_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.SurveyConduct, "respondent_type_concept_id"
    )
    timing_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.SurveyConduct, "timing_concept_id"
    )
    collection_method_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.SurveyConduct, "collection_method_concept_id"
    )
    assisted_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.SurveyConduct, "assisted_source_value"
    )
    respondent_type_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.SurveyConduct, "respondent_type_source_value"
    )
    timing_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.SurveyConduct, "timing_source_value"
    )
    collection_method_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.SurveyConduct, "collection_method_source_value"
    )
    survey_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.SurveyConduct, "survey_source_value"
    )
    survey_source_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.SurveyConduct, "survey_source_concept_id"
    )
    survey_source_identifier: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.SurveyConduct, "survey_source_identifier"
    )
    validated_survey_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.SurveyConduct, "validated_survey_concept_id"
    )
    validated_survey_source_value: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.SurveyConduct, "validated_survey_source_value"
    )
    survey_version_number: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.SurveyConduct, "survey_version_number"
    )
    visit_occurrence_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.SurveyConduct, "visit_occurrence_id"
    )
    response_visit_occurrence_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.SurveyConduct, "response_visit_occurrence_id"
    )


class FactRelationship(Base, NoIdRowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.FactRelationship)

    domain_concept_id_1: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.FactRelationship, "domain_concept_id_1"
    )
    fact_id_1: Mapped[int] = create_mapped_column(
        DOMAIN, model.FactRelationship, "fact_id_1"
    )
    domain_concept_id_2: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.FactRelationship, "domain_concept_id_2"
    )
    fact_id_2: Mapped[int] = create_mapped_column(
        DOMAIN, model.FactRelationship, "fact_id_2"
    )
    relationship_concept_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.FactRelationship, "relationship_concept_id"
    )
    fact_relationship_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.FactRelationship, "fact_relationship_id"
    )


class MeasurementRelation(Base, NoIdRowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.MeasurementRelation)

    measurement_relation_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.MeasurementRelation, "measurement_relation_id"
    )
    from_measurement_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.MeasurementRelation, "from_measurement_id"
    )
    to_measurement_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.MeasurementRelation, "to_measurement_id"
    )



# pylint: disable=too-few-public-methods
# This module defines base classes, methods are added later


from typing import ClassVar

from gen_epix.commondb.domain.command import CrudCommand
from gen_epix.omopdb.domain import model


class PersonCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Person


class ObservationPeriodCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.ObservationPeriod


class VisitOccurrenceCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.VisitOccurrence


class VisitDetailCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.VisitDetail


class ConditionOccurrenceCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.ConditionOccurrence


class DrugExposureCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.DrugExposure


class ProcedureOccurrenceCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.ProcedureOccurrence


class DeviceExposureCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.DeviceExposure


class MeasurementCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Measurement


class ObservationCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Observation


class NoteCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Note


class NoteNlpCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.NoteNlp


class SpecimenCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Specimen


class FactRelationshipCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.FactRelationship


class SurveyConductCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.SurveyConduct


class LocationCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Location


class LocationHistoryCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.LocationHistory


class CareSiteCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.CareSite


class ProviderCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Provider


class PayerPlanPeriodCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.PayerPlanPeriod


class CostCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Cost


class DrugEraCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.DrugEra


class DoseEraCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.DoseEra


class ConditionEraCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.ConditionEra


class MetadataCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Metadata


class CdmSourceCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.CdmSource


class ConceptCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Concept


class VocabularyCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Vocabulary


class DomainCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Domain


class ConceptClassCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.ConceptClass


class ConceptRelationshipCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.ConceptRelationship


class RelationshipCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Relationship


class ConceptSynonymCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.ConceptSynonym


class ConceptAncestorCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.ConceptAncestor


class SourceToConceptMapCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.SourceToConceptMap


class DrugStrengthCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.DrugStrength


class CohortCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Cohort


class CohortDefinitionCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.CohortDefinition


class MeasurementRelationCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.MeasurementRelation



"""Refactored configuration management using Strategy Pattern."""

import abc
import copy
import importlib
import logging
import logging.config as logging_config
import os
from enum import Enum
from locale import getpreferredencoding
from pathlib import Path

import yaml  # type: ignore[import-untyped]
from dynaconf import Dynaconf

from gen_epix.commondb.config.factory import IdFactory, TimestampFactory
from gen_epix.commondb.config.settings_manager import SettingsManager
from gen_epix.fastapp import App


class BaseAppCfg(abc.ABC):
    """Abstract base class for application configuration."""

    def __init__(self) -> None:
        self._name: str | None
        self._app_name: str
        self._service_type_enum: type[Enum]
        self._repository_type_enum: type[Enum]
        self._log_setup: bool
        self._cfg: Dynaconf
        self._setup_logger: logging.Logger
        self._api_logger: logging.Logger
        self._app_logger: logging.Logger
        self._service_logger: logging.Logger

    @property
    def name(self) -> str:
        if self._name is None:
            raise ValueError("name is not set")
        return self._name

    @property
    def app_name(self) -> str:
        return self._app_name

    @property
    def service_type_enum(self) -> type[Enum]:
        return self._service_type_enum

    @property
    def repository_type_enum(self) -> type[Enum]:
        return self._repository_type_enum

    @property
    def log_setup(self) -> bool:
        return self._log_setup

    @property
    def cfg(self) -> Dynaconf:
        return self._cfg

    @property
    def setup_logger(self) -> logging.Logger:
        return self._setup_logger

    @property
    def api_logger(self) -> logging.Logger:
        return self._api_logger

    @property
    def app_logger(self) -> logging.Logger:
        return self._app_logger

    @property
    def service_logger(self) -> logging.Logger:
        return self._service_logger

    @abc.abstractmethod
    def copy_repository_files(
        self,
        tgt_dir: Path | str,
        service_type: Enum | None = None,
        on_exist: str = "skip",
    ) -> None:
        raise NotImplementedError("Subclasses must implement copy_repository_files")


class AppCfg(BaseAppCfg):
    """Main application configuration class using Strategy Pattern."""

    @staticmethod
    def _prefix_envvar(
        envvar_prefix: str | None, envvar: str, delimiter: str = "_"
    ) -> str:
        """Create prefixed environment variable name."""
        if envvar_prefix:
            return f"{envvar_prefix}{delimiter}{envvar}"
        return envvar

    @staticmethod
    def _prefix_logger(
        logger_prefix: str | None, logger_name: str, delimiter: str = "."
    ) -> str:
        """Create prefixed logger name."""
        if logger_prefix:
            return f"{logger_prefix}{delimiter}{logger_name}"
        return logger_name

    def __init__(
        self,
        app_name_or_enum: Enum | str,
        service_type_enum: type[Enum],
        repository_type_enum: type[Enum],
        log_setup: bool = True,
        name: str | None = None,
        setup_logger_level: str | int | None = None,
        logger_prefix: str | None = None,
        envvar_prefix: str | None = None,
        log_config_file_envvar: str = "LOG_CONFIG_FILE",
        log_level_envvar: str = "LOG_LEVEL",
    ):
        """Initialize application configuration.

        Args:
            app_name: Name of the application (e.g., 'commondb', 'casedb')
            service_type_enum: Enum of service types
            repository_type_enum: Enum of repository types
            name: optional name for this configuration instance that can e.g. be used
               as a key in collection of configurations
            log_setup: Whether to set up logging
            logger_prefix: Prefix for logger names (defaults to app_name.lower())
            envvar_prefix: Prefix for environment variables (defaults to app_name.upper())
            logging_config_file_envvar: Environment variable for logging config file
            idps_config_file_envvar: Environment variable for identity provider config
            logging_level_from_secret_envvar: Environment variable to control log level from secrets
        """
        # Parse input
        if isinstance(app_name_or_enum, Enum):
            app_name = str(app_name_or_enum.value)
        else:
            app_name = app_name_or_enum

        # Add some properties
        self._app_name = app_name
        self._name = name
        self._service_type_enum = service_type_enum
        self._repository_type_enum = repository_type_enum
        self._envvar_prefix = envvar_prefix or f"{app_name.upper()}_"
        self._logger_prefix = logger_prefix or app_name.lower()
        self._log_config_file_envvar = log_config_file_envvar
        self._log_setup = log_setup
        self._log_level_envvar = log_level_envvar
        self._setup_logger_level = setup_logger_level

        # Configure and set loggers
        self._init_configure_loggers()
        self.set_log_level()
        if self._setup_logger_level is not None:
            self.setup_logger.setLevel(self._setup_logger_level)
        if self._log_setup:
            self.setup_logger.info(
                App.create_static_log_message("c6010f14", "Started loading config data")
            )

        # Load settings
        self._init_load_settings()
        self.set_log_level()
        if self._log_setup:
            self.setup_logger.info(
                App.create_static_log_message(
                    "a7b3c4d5", f"Loaded settings from {type(self._cfg).__name__}"
                )
            )

        # Validate settings
        self._init_validate_settings()
        if self._log_setup:
            self.setup_logger.info(
                App.create_static_log_message(
                    "cdb7abcb", "Finished loading config data"
                )
            )

    def _init_configure_loggers(
        self,
    ) -> None:
        """Configure loggers from logging configuration file."""
        logging_config_file = os.environ[
            f"{self._envvar_prefix}{self._log_config_file_envvar}"
        ]
        with open(logging_config_file, "rt", encoding=getpreferredencoding()) as handle:
            logging_config_yaml = yaml.safe_load(handle.read())
            logging_config.dictConfig(logging_config_yaml)

        # Get loggers and put as attributes
        self._setup_logger = logging.getLogger(
            AppCfg._prefix_logger(self._logger_prefix, "setup")
        )
        self._api_logger = logging.getLogger(
            AppCfg._prefix_logger(self._logger_prefix, "api")
        )
        self._app_logger = logging.getLogger(
            AppCfg._prefix_logger(self._logger_prefix, "app")
        )
        self._service_logger = logging.getLogger(
            AppCfg._prefix_logger(self._logger_prefix, "service")
        )
        self._logging_config_yaml = logging_config_yaml

    def _init_load_settings(self) -> None:
        """Load settings using SettingsManager."""

        settings_manager = SettingsManager(prefix=self._envvar_prefix)
        self._cfg = settings_manager.load_settings()

    def _init_validate_settings(self) -> None:
        """Validate settings and apply defaults to all services and repositories."""
        # Map timestamp and id factories
        defaults_cfg = self._cfg["service"]["defaults"]["props"]
        timestamp_factory = getattr(
            TimestampFactory,
            defaults_cfg["timestamp_factory"],
        )
        id_factory = getattr(IdFactory, defaults_cfg["id_factory"])
        defaults_cfg["timestamp_factory"] = timestamp_factory
        defaults_cfg["id_factory"] = id_factory

        # Map default repository type
        repository_type = getattr(
            self._repository_type_enum, self._cfg["repository"]["defaults"]["type"]
        )
        self._cfg["repository"]["defaults"]["type"] = repository_type

        # Get class for and apply defaults to each service and repository
        for service_type in self._service_type_enum:
            service_type_str = service_type.value.lower()
            if service_type_str not in self._cfg["service"]:
                self._cfg["service"].update({service_type_str: {}})
            service_cfg = self._cfg["service"][service_type_str]

            # Get class for service
            service_module = service_cfg["module"]
            service_class_name = service_cfg["class_name"]
            service_cfg["class"] = getattr(
                importlib.import_module(service_module), service_class_name
            )

            # Apply defaults to service
            orig_cfg = copy.deepcopy(service_cfg)
            service_cfg.update(self._cfg["service"]["defaults"])
            service_cfg.update(orig_cfg)

            # Skip if the service does not have a repository
            if service_type_str not in self._cfg["repository"]:
                continue
            repository_cfg = self._cfg["repository"][service_type_str]

            # Get class for repository
            repository_module = repository_cfg["module"]
            repository_class_name = repository_cfg["class_name"]
            repository_cfg["class"] = getattr(
                importlib.import_module(repository_module), repository_class_name
            )

            # Apply defaults to repository, if the service has a repository
            orig_cfg = copy.deepcopy(repository_cfg)
            repository_cfg.update(self._cfg["repository"]["defaults"])
            repository_cfg.update(orig_cfg)

    def copy_repository_files(
        self,
        tgt_dir: Path | str,
        service_type: Enum | None = None,
        on_exist: str = "skip",
    ) -> None:
        """
        Copy any repository files to a new folder and update the configuration
        correspondingly. This is useful e.g. for creating isolated test environments.
        """
        # Parse input
        if isinstance(tgt_dir, str):
            tgt_dir = Path(tgt_dir)
        if not tgt_dir.is_dir():
            raise ValueError(f"new_folder {tgt_dir} is not a directory")
        if on_exist not in ("skip", "overwrite", "raise"):
            raise ValueError(
                f"on_exist must be 'skip', 'overwrite' or 'raise', got {on_exist}"
            )
        if service_type is None:
            service_types = list(self.service_type_enum)
        else:
            service_types = [service_type]
        # Go over each service type
        for service_type in service_types:
            # Get file path from config
            service_type_str = service_type.value.lower()
            if service_type_str not in self._cfg["repository"]:
                continue
            cfg = self._cfg["repository"][service_type_str]["props"]
            if "file" not in cfg:
                continue
            curr_path = Path(cfg["file"])
            new_path = tgt_dir / curr_path.name
            # Copy file
            if not curr_path.exists():
                raise FileNotFoundError(f"Source file not found: {curr_path}")
            if new_path.exists():
                if on_exist == "overwrite":
                    pass
                elif on_exist == "skip":
                    continue
                elif on_exist == "raise":
                    raise FileExistsError(
                        f"Destination file already exists: {new_path}"
                    )
                else:
                    raise NotImplementedError(
                        f"on_exist value '{on_exist}' not implemented"
                    )
            with open(curr_path, "rb") as src_handle:
                with open(new_path, "wb") as dst_handle:
                    dst_handle.write(src_handle.read())
            # Update config
            cfg["file"] = str(Path(tgt_dir) / curr_path.name)

    def set_log_level(self, log_level: str | int | None = None) -> None:
        """Set log level for all loggers."""
        # Parse log level
        log_level_envvar = f"{self._envvar_prefix}{self._log_level_envvar}"
        if log_level is None:
            if log_level_envvar in os.environ:
                # Get log level from environment variable, before settings
                log_level = os.environ[log_level_envvar]
            elif hasattr(self, "_cfg"):
                # Get log level from settings, if available
                log_level = self._cfg["log"]["level"]
        if log_level is None:
            # No log level available
            return
        if isinstance(log_level, str):
            log_level = log_level.upper()
        # Set new log level for all in settings as well
        if hasattr(self, "_cfg"):
            self._cfg["log"]["level"] = log_level
        for handler in self._setup_logger.handlers:
            handler.setLevel(log_level)
        self._setup_logger.setLevel(log_level)
        for logger_name in self._logging_config_yaml["loggers"]:
            curr_logger = logging.getLogger(logger_name)
            if self._log_setup:
                self.setup_logger.debug(
                    App.create_static_log_message(
                        "6ba9367c",
                        f"Updated logger {logger_name} with level {log_level}",
                    )
                )
            for handler in curr_logger.handlers:
                handler.setLevel(log_level)
            curr_logger.setLevel(log_level)



import os
from collections.abc import Callable

from fastapi import FastAPI, Request, Response
from starlette.middleware.base import BaseHTTPMiddleware

from gen_epix.commondb.util import get_package_version


class UpdateResponseHeaderMiddleware(BaseHTTPMiddleware):
    def __init__(
        self,
        app: FastAPI,
        general_headers: dict[str, str] | None = None,
        exception_headers: list[tuple[set[str], dict[str, str]]] | None = None,
    ):
        super().__init__(app)
        self._general_headers = general_headers or {}
        self._exception_headers = exception_headers

    async def dispatch(self, request: Request, call_next: Callable) -> Response:
        response: Response = await call_next(request)
        if not self._exception_headers:
            if self._general_headers:
                response.headers.update(self._general_headers)
            response.headers.update(self._general_headers)
            return response
        for endpoints, headers in self._exception_headers:
            if request["path"] in endpoints:
                response.headers.update(headers)
                return response
        response.headers.update(self._general_headers)
        app_version = os.environ.get("APP_VERSION")
        if app_version:
            content_type = response.headers.get("content-type")
            if content_type:
                content_type = content_type + ";version=" + app_version
                response.headers.update({"content-type": content_type})
        response.headers["API-Version"] = get_package_version()
        return response



from enum import Enum
from typing import TypeVar, overload

from pydantic import BaseModel, ConfigDict, Field, computed_field, field_validator

from gen_epix import fastapp
from gen_epix.commondb.domain import command, enum, model
from gen_epix.fastapp import BaseUserManager
from gen_epix.fastapp.services.rbac import BaseRbacService

# TypeVars for get_mapped_class method to express subclass relationships
ModelType = TypeVar("ModelType", bound=model.Model)
CommandType = TypeVar("CommandType", bound=command.Command)
PolicyType = TypeVar("PolicyType", bound=fastapp.Policy)


class AppImplDetails(BaseModel):
    """
    Implementation details for the App.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    sorted_service_types: list[Enum] = Field(
        description="List of service types in DAG sorted order for initialization"
    )
    services: dict[Enum, fastapp.BaseService] = Field(
        default_factory=dict, description="Dictionary of services keyed by service type"
    )
    repositories: dict[Enum, fastapp.BaseRepository] = Field(
        default_factory=dict,
        description="Dictionary of repositories keyed by service type",
    )
    registered_user_dependency_or_none: model.User | None = Field(
        default=None,
        description="Dependency that provides the currently registered user",
    )
    new_user_dependency_or_none: model.User | None = Field(
        default=None, description="Dependency that provides a new user"
    )
    idp_user_dependency_or_none: model.User | None = Field(
        default=None,
        description="Dependency that provides a user known by the identity provider but not registered in the application",
    )
    model_class_map: dict[type[model.Model], type[model.Model]] = Field(
        default_factory=dict,
        description="Mapping of commondb model classes to any derived implementing model classes",
    )
    command_class_map: dict[type[command.Command], type[command.Command]] = Field(
        default_factory=dict,
        description="Mapping of commondb command classes to any derived implementing command classes",
    )
    policy_class_map: dict[type[fastapp.Policy], type[fastapp.Policy]] = Field(
        default_factory=dict,
        description="Mapping of commondb policy classes to any derived implementing policy classes",
    )
    rbac_service_class: type[BaseRbacService] = Field(
        description="Class used for RBAC service"
    )
    user_manager_class: type[BaseUserManager] = Field(
        description="Class used for user management"
    )
    role_map: dict[enum.Role | Enum, str] = Field(
        description="Mapping of roles to their string representations. Where possible, the key is the commondb role enum.",
    )
    role_set_map: dict[enum.RoleSet | Enum, frozenset[str]] = Field(
        description="Mapping of role sets to their corresponding sets of role strings. Where possible, the key is the commondb role set enum.",
    )
    role_permissions_map: dict[
        str, set[tuple[type[fastapp.Command], fastapp.PermissionType]]
    ] = Field(
        description="Mapping of roles as strings to (command, permission type) tuples",
    )

    @computed_field
    def rev_role_map(self) -> dict[str, enum.Role | Enum]:
        return {x: y for y, x in self.role_map.items()}

    @computed_field
    def registered_user_dependency(self) -> model.User:
        if self.registered_user_dependency_or_none is None:
            raise ValueError("registered_user_dependency is not set")
        return self.registered_user_dependency_or_none

    @computed_field
    def new_user_dependency(self) -> model.User:
        if self.new_user_dependency_or_none is None:
            raise ValueError("new_user_dependency is not set")
        return self.new_user_dependency_or_none

    @computed_field
    def idp_user_dependency(self) -> model.User:
        if self.idp_user_dependency_or_none is None:
            raise ValueError("idp_user_dependency is not set")
        return self.idp_user_dependency_or_none

    @field_validator("role_map", mode="before")
    @classmethod
    def _validate_role_map(
        cls, value: dict[enum.Role | Enum, str] | type[Enum]
    ) -> dict[enum.Role | Enum, str]:
        if not isinstance(value, dict):
            value = {e: str(e.value) for e in value}
            if len(set(value.values())) != len(value):
                raise ValueError("role_map must not contain duplicate values")
        return value

    @field_validator("role_set_map", mode="before")
    @classmethod
    def _validate_role_set_map(
        cls, value: dict[enum.RoleSet | Enum, str] | type[Enum]
    ) -> dict[enum.RoleSet | Enum, str]:
        if not isinstance(value, dict):
            # Allow passing Enum type directly for convenience
            value = {e: e.value for e in value}
        return value

    @field_validator("sorted_service_types", mode="before")
    @classmethod
    def _validate_sorted_service_types(
        cls, value: tuple[Enum, ...] | list[Enum]
    ) -> list[Enum]:
        if isinstance(value, tuple):
            return list(value)
        if len(set(value)) != len(value):
            raise ValueError("sorted_service_types must not contain duplicates")
        return value

    @field_validator(
        "model_class_map", "command_class_map", "policy_class_map", mode="after"
    )
    @classmethod
    def _validate_model_class_map(cls, value: dict[type, type]) -> dict[type, type]:
        for x, y in value.items():
            if not issubclass(y, x):
                raise ValueError(f"Class {y} is not a subclass of {x}")
        return value

    @overload
    def get_mapped_class(self, base_class: type[ModelType]) -> type[ModelType]: ...

    @overload
    def get_mapped_class(self, base_class: type[CommandType]) -> type[CommandType]: ...

    @overload
    def get_mapped_class(self, base_class: type[PolicyType]) -> type[PolicyType]: ...

    def get_mapped_class(
        self,
        base_class: type[model.Model] | type[command.Command] | type[fastapp.Policy],
    ) -> type[model.Model] | type[command.Command] | type[fastapp.Policy]:
        """
        Get the mapped class for a given base class, or return the base class if no mapping exists.
        """
        if issubclass(base_class, model.Model):
            return self.model_class_map.get(base_class, base_class)
        if issubclass(base_class, command.Command):
            return self.command_class_map.get(base_class, base_class)
        if issubclass(base_class, fastapp.Policy):
            return self.policy_class_map.get(base_class, base_class)
        raise ValueError(f"Unsupported base class type: {base_class}")



from typing import Any
from uuid import UUID

from gen_epix.casedb.domain import command, model
from gen_epix.commondb.domain.service import BaseAbacService
from gen_epix.commondb.policies import (
    IsOrganizationAdminPolicy as CommonIsOrganizationAdminPolicy,
)
from gen_epix.fastapp import CrudOperation


class IsOrganizationAdminPolicy(CommonIsOrganizationAdminPolicy):

    def __init__(
        self,
        abac_service: BaseAbacService,
        **kwargs: Any,
    ):
        super().__init__(
            abac_service,
            **kwargs,
        )
        f = self.register_retrieve_organization_ids_handler
        f(
            command.OrganizationAccessCasePolicyCrudCommand,
            self._get_organization_ids_for_organization_case_policy,
        )
        f(
            command.OrganizationShareCasePolicyCrudCommand,
            self._get_organization_ids_for_organization_case_policy,
        )
        f(
            command.UserAccessCasePolicyCrudCommand,
            self._get_organization_ids_for_user_case_policy,
        )
        f(
            command.UserShareCasePolicyCrudCommand,
            self._get_organization_ids_for_user_case_policy,
        )

    def _get_organization_ids_for_organization_case_policy(
        self,
        cmd: (
            command.OrganizationAccessCasePolicyCrudCommand
            | command.OrganizationShareCasePolicyCrudCommand
        ),
    ) -> set[UUID]:
        policies: list[model.OrganizationAccessCasePolicy] | list[model.OrganizationShareCasePolicy] = cmd.get_objs()  # type: ignore[assignment]
        return {x.organization_id for x in policies}

    def _get_organization_ids_for_user_case_policy(
        self,
        cmd: (
            command.UserAccessCasePolicyCrudCommand
            | command.UserShareCasePolicyCrudCommand
        ),
    ) -> set[UUID]:
        policies: list[model.UserAccessCasePolicy] | list[model.UserShareCasePolicy] = cmd.get_objs()  # type: ignore[assignment]
        users: list[model.User] = self.abac_service.app.handle(
            command.UserCrudCommand(
                user=cmd.user,
                objs=None,
                obj_ids=list(set(x.user_id for x in policies)),
                operation=CrudOperation.READ_SOME,
            )
        )
        return {x.organization_id for x in users}



"""
Tuple map transformer implementation.
"""

from collections.abc import Hashable
from typing import Any

from gen_epix.transform.adapter import ObjectAdapter
from gen_epix.transform.transformer import Transformer


class TupleMapTransformer(Transformer):
    """
    Maps tuples of values from source keys to target keys.
    """

    def __init__(
        self,
        map_rows: list[dict[Hashable, Any]],
        src_fields: list[Hashable],
        tgt_fields: list[Hashable],
        name: str | None = None,
        map_fields: list[Hashable] | None = None,
        is_active_map_field: Hashable | None = None,
    ) -> None:
        # Verify input
        if len(set(src_fields)) < len(src_fields):
            raise ValueError("Source column names must be unique")
        if len(set(tgt_fields)) < len(tgt_fields):
            raise ValueError("Target column names must be unique")
        map_fields = map_fields or []
        if len(set(map_fields)) < len(map_fields):
            raise ValueError("Map column names must be unique")

        # Initialise some
        super().__init__(name)
        self._src_fields = src_fields
        self._tgt_fields = tgt_fields
        self._map_fields = map_fields
        self._all_fields = src_fields + tgt_fields + map_fields
        if len(set(self._all_fields)) < len(self._all_fields):
            raise ValueError("All column names together must be unique")
        self._is_active_map_field = is_active_map_field

    def update_map(self, map_df: list[dict[Hashable, Any]]) -> None:
        tuple_map: dict[tuple, tuple] = {}
        # Extract source and target tuples
        for row in map_df:
            key = tuple(row.get(x) for x in self._map_fields)
            if key in tuple_map:
                raise ValueError(f"Duplicate mapping for map field {key}")
            if self._is_active_map_field is not None and not row.get(
                self._is_active_map_field, True
            ):
                # Skip inactive mapping
                continue
            value = tuple(row.get(x) for x in self._tgt_fields)
            tuple_map[key] = value
        self._map_df = map_df
        self._tuple_map = tuple_map

    def transform(self, obj: ObjectAdapter) -> ObjectAdapter:
        """Map a Transform all specified fields."""
        key = tuple(obj.get(x) for x in self._src_fields)
        if key not in self._tuple_map:
            raise ValueError(f"Validation failed for object: {obj.unwrap()}")
        values = self._tuple_map[key]
        for field, value in zip(self._tgt_fields, values):
            obj.set(field, value)
        return obj



# pylint: disable=useless-import-alias
from gen_epix.commondb.api import (
    UpdateUserOwnOrganizationRequestBody as UpdateUserOwnOrganizationRequestBody,
)
from gen_epix.seqdb.api.seq import (
    RetrieveAlleleProfileRequestBody as RetrieveAlleleProfileRequestBody,
)
from gen_epix.seqdb.api.seq import (
    RetrievePhylogeneticTreeRequestBody as RetrievePhylogeneticTreeRequestBody,
)
from gen_epix.seqdb.api.seq import RetrieveSeqRequestBody as RetrieveSeqRequestBody



import datetime
import gzip
import json
import pickle
import zipfile
from collections.abc import Callable, Hashable, Iterable
from functools import partial
from pathlib import Path
from typing import Any
from uuid import UUID

from gen_epix.fastapp import exc
from gen_epix.fastapp.domain.entity import Entity
from gen_epix.fastapp.enum import CrudOperation, FieldTypeSet
from gen_epix.fastapp.model import Model
from gen_epix.fastapp.repositories.dict.unit_of_work import DictUnitOfWork
from gen_epix.fastapp.repository import BaseRepository
from gen_epix.fastapp.unit_of_work import BaseUnitOfWork
from gen_epix.filter import CompositeFilter, Filter, LogicalOperator


class DictRepository(BaseRepository):

    @classmethod
    def create_repository(cls, **kwargs: Any) -> BaseRepository:
        entities = kwargs.pop("entities", [])
        file = kwargs.pop("file", None)
        file_type = kwargs.pop("file_type", None)
        if file is None:
            raise exc.RepositoryInitializationServiceError("No file provided")
        if file_type is None:
            path = Path(file)
            suffixes = [x.lower() for x in path.suffixes]
            if ".pkl" in suffixes:
                file_type = "pkl"
            elif ".json" in suffixes:
                file_type = "json"
            elif ".zip" in suffixes:
                file_type = "zip"
        file_type = file_type.lower()
        if file_type == "pkl":
            return DictRepository.create_repository_from_pkl(
                repository_class=cls,
                entities=entities,
                pkl_file=file,
                **kwargs,
            )
        elif file_type == "zip":
            return DictRepository.create_repository_from_json(
                repository_class=cls,
                entities=entities,
                zip_file=file,
                **kwargs,
            )
        else:
            raise NotImplementedError(f"Unsupported file type: {file_type}")

    @staticmethod
    def create_repository_from_pkl(
        repository_class: type[BaseRepository],
        entities: Iterable[Entity],
        pkl_file: str,
        **kwargs: Any,
    ) -> "DictRepository":
        if pkl_file.lower().endswith(".gz"):
            with gzip.open(pkl_file, "rb") as handle:
                db = pickle.load(handle)
        else:
            with open(pkl_file, "rb") as handle:
                db = pickle.load(handle)
        # TODO: check validity of db
        repository = repository_class(entities, db, **kwargs)  # type: ignore[call-arg]
        assert isinstance(repository, DictRepository)
        return repository

    @staticmethod
    def create_repository_from_json(
        repository_class: type[BaseRepository],
        entities: Iterable[Entity],
        zip_file: str,
        **kwargs: Any,
    ) -> "DictRepository":
        if not zip_file.lower().endswith(".zip"):
            raise exc.RepositoryServiceError("Invalid file format. Expected .zip")
        db = {}
        entities = list(entities)
        with zipfile.ZipFile(zip_file, "r") as zip_handle:
            files = set(zip_handle.namelist())
            for entity in entities:
                if not entity.persistable:
                    continue
                json_file = entity.name + ".json"
                if json_file not in files and entity.table_name:
                    json_file = entity.table_name + ".json"
                if json_file not in files:
                    raise exc.RepositoryServiceError(
                        f"Missing file for entity {entity.name} in archive {zip_file}"
                    )
                model_class = entity.model_class
                with zip_handle.open(json_file) as handle:
                    id_field_name = entity.id_field_name
                    db[model_class] = {
                        getattr(y, id_field_name): y  # type: ignore[arg-type]
                        for y in (model_class(**x) for x in json.load(handle))
                    }
        repository = repository_class(entities, db, **kwargs)  # type: ignore[call-arg]
        assert isinstance(repository, DictRepository)
        return repository

    def __init__(
        self,
        entities: Iterable[Entity],
        db: dict[type[Model], dict[Hashable, Model]],
        **kwargs: Any,
    ):
        extra_data = kwargs.pop("extra_data", "ignore")
        missing_data = kwargs.pop("missing_data", "raise")
        timestamp_factory = kwargs.pop("timestamp_factory", datetime.datetime.now)
        if extra_data not in {"ignore", "raise", "drop"}:
            raise ValueError(f"Invalid extra_data: {extra_data}")
        if missing_data not in {"raise", "ignore"}:
            raise ValueError(f"Invalid missing_data: {missing_data}")
        # Initialize properties
        self._db = dict(db.items())
        self._timestamp_factory = timestamp_factory
        self._entities = set(entities)
        self._links: dict[
            type[Model],
            list[tuple[str, type[Model], str, int, dict[Hashable, Model] | None]],
        ] = {}
        self._get_id: dict[type[Model], Callable[[Model], Hashable]] = {}
        self._back_links: dict[type[Model], list[tuple[type[Model], str]]] = {}
        self._value_field_names: dict[type[Model], list[str]] = {}
        self._keys_generators: dict[type[Model], dict[int, Callable[[Model], str]]] = {}
        self._init_properties(entities, db, missing_data)

        self._verify_extra_models_and_extract_reverse_links(extra_data)

    @property
    def db(self) -> dict[type[Model], dict[Hashable, Model]]:
        return self._db

    def _init_properties(
        self, entities: Iterable[Entity], db: dict, missing_data: str
    ) -> None:
        # Further populate properties
        for entity in entities:
            # Extract entity data
            model_class = entity.model_class
            assert issubclass(model_class, Model)
            self._links[model_class] = self._get_links(entity)
            self._back_links[model_class] = []
            self._value_field_names[model_class] = []
            for field_type in FieldTypeSet.DATA.value:
                self._value_field_names[model_class].extend(
                    entity.get_field_names(field_type=field_type)
                )
            self._keys_generators[model_class] = entity.get_keys_generator()  # type: ignore[assignment]
            if entity.persistable and model_class not in db:
                if missing_data == "ignore":
                    self._db[model_class] = {}
                elif missing_data == "raise":
                    raise ValueError(f"No data for model {model_class}")
                else:
                    raise NotImplementedError
            # Create ID getter
            id_field_name = entity.id_field_name
            if id_field_name is None:
                # No ID field defined for this model, no getter can be created
                continue
            if not isinstance(id_field_name, str):
                raise NotImplementedError(
                    f"Model {model_class.__name__} has more than one ID field"
                )
            self._get_id[model_class] = partial(
                lambda x, y: getattr(y, x), id_field_name
            )

    def _verify_extra_models_and_extract_reverse_links(self, extra_data: str) -> None:
        # Verify extra Models in db and extract reverse links
        for model_class, links in self._links.items():
            to_pop = []
            for i, link in enumerate(links):
                link_field_name, link_model_class, _, _, _ = link
                if link_model_class not in self._links:
                    if extra_data == "ignore":
                        continue
                    if extra_data == "drop":
                        to_pop.append(i)
                        continue
                    if extra_data == "raise":
                        raise ValueError(
                            f"Model {model_class} links to "
                            "additional linked model {link_model_class}"
                        )
                    raise NotImplementedError
                self._back_links[link_model_class].append(
                    (model_class, link_field_name)
                )
            for i in reversed(to_pop):
                links.pop(i)

    def crud(
        self,
        uow: BaseUnitOfWork,
        user_id: Hashable | None,
        model_class: type[Model],
        objs: Model | Iterable[Model] | None,
        obj_ids: Hashable | Iterable[Hashable] | None,
        operation: CrudOperation,
        filter: Filter | None = None,
        **kwargs: Any,
    ) -> Hashable | list[Hashable] | Model | list[Model] | bool | list[bool] | None:
        BaseRepository.verify_crud_args(model_class, objs, obj_ids, operation)
        match operation:
            case CrudOperation.READ_ALL:
                return self.read_all(model_class, filter, **kwargs)
            case CrudOperation.READ_ONE:
                return self.read_one(model_class, obj_ids, **kwargs)  # type: ignore[arg-type]
            case CrudOperation.READ_SOME:
                return self.read_some(model_class, obj_ids, **kwargs)  # type: ignore[arg-type]
            case CrudOperation.CREATE_ONE:
                return self.upsert_some(
                    user_id,
                    model_class,
                    objs,  # type: ignore[arg-type]
                    raise_on_present=True,
                    raise_on_missing=False,
                    **kwargs,
                )
            case CrudOperation.CREATE_SOME:
                return self.upsert_some(
                    user_id,
                    model_class,
                    objs,  # type: ignore[arg-type]
                    raise_on_present=True,
                    raise_on_missing=False,
                    **kwargs,
                )
            case CrudOperation.UPDATE_ONE:
                return self.upsert_some(
                    user_id,
                    model_class,
                    objs,  # type: ignore[arg-type]
                    raise_on_present=False,
                    raise_on_missing=True,
                    **kwargs,
                )
            case CrudOperation.UPDATE_SOME:
                return self.upsert_some(
                    user_id,
                    model_class,
                    objs,  # type: ignore[arg-type]
                    raise_on_present=False,
                    raise_on_missing=True,
                    **kwargs,
                )
            case CrudOperation.UPSERT_ONE:
                return self.upsert_some(
                    user_id,
                    model_class,
                    objs,  # type: ignore[arg-type]
                    raise_on_present=False,
                    raise_on_missing=False,
                    **kwargs,
                )
            case CrudOperation.UPSERT_SOME:
                return self.upsert_some(
                    user_id,
                    model_class,
                    objs,  # type: ignore[arg-type]
                    raise_on_present=False,
                    raise_on_missing=False,
                    **kwargs,
                )
            case CrudOperation.DELETE_ONE:
                return self.delete_some(model_class, obj_ids, **kwargs)
            case CrudOperation.DELETE_SOME:
                return self.delete_some(model_class, obj_ids, **kwargs)
            case CrudOperation.DELETE_ALL:
                return self.delete_all(model_class, filter, **kwargs)
            case CrudOperation.EXISTS_ONE:
                return self.exists_one(model_class, obj_ids)  # type: ignore[arg-type]
            case CrudOperation.EXISTS_SOME:
                return self.exists_some(model_class, obj_ids)  # type: ignore[arg-type]
            case _:
                raise NotImplementedError(f"Operation {operation} not implemented")

    def read_fields(
        self,
        uow: BaseUnitOfWork,
        user_id: Hashable | None,
        model_class: type[Model],
        field_names: list[str],
        filter: Filter | None = None,
        **kwargs: Any,
    ) -> Iterable[tuple]:
        all_objs_iterable = self._db[model_class].values()
        if filter:
            for obj in filter.filter_rows(all_objs_iterable, is_model=True):
                yield tuple(getattr(obj, x) for x in field_names)
        else:
            for obj in all_objs_iterable:
                yield tuple(getattr(obj, x) for x in field_names)

    def uow(self, **kwargs: Any) -> BaseUnitOfWork:
        return DictUnitOfWork()

    def split_filter(
        self, model_class: type, filter: Filter | None
    ) -> tuple[Filter | None, Filter | None]:
        # Entire filter can be used as where clause since the data are stored as
        # domain models
        return filter, None

    def verify_valid_ids(
        self,
        uow: BaseUnitOfWork,
        user_id: Hashable,
        model_class: type[Model],
        obj_ids: Iterable[Hashable],
        verify_exists: bool = True,
        verify_duplicate: bool = True,
    ) -> None:
        if verify_exists:
            df = self._db[model_class]
            invalid_obj_ids = [x for x in obj_ids if x not in df]
            if invalid_obj_ids:
                DictRepository._raise_invalid_ids(model_class, invalid_obj_ids)
        if verify_duplicate:
            DictRepository._verify_duplicate_ids(model_class, obj_ids)

    def read_all(
        self,
        model_class: type[Model],
        filter: Filter | None,
        cascade_read: bool = False,
        return_id: bool = False,
        **kwargs: Any,
    ) -> list[Model] | list[Hashable]:
        return_copy = kwargs.get("return_copy", True)
        df = self._db[model_class]
        # Get any query filter
        query_filter: Filter | None = None
        obj_filter: Filter | None = kwargs.get("obj_filter")
        if filter and obj_filter:
            query_filter = CompositeFilter(
                filters=[filter, obj_filter], operator=LogicalOperator.AND  # type: ignore[list-item]
            )
        elif filter:
            query_filter = filter
        elif obj_filter:
            query_filter = obj_filter
        else:
            query_filter = None
        # Get matching objects
        objs: list[Model] | list[Hashable]
        if query_filter:
            if return_id:
                objs = [
                    x
                    for x, y in zip(
                        df.keys(), query_filter.match_rows(df.values(), is_model=True)
                    )
                    if y
                ]
            else:
                objs = list(query_filter.filter_rows(df.values(), is_model=True))  # type: ignore[assignment]
        elif return_id:
            objs = list(df.keys())
        else:
            objs = list(df.values())
        # Make copy of objects for returning if necessary
        if not return_id and return_copy:
            objs = [x.model_copy() for x in objs if x]  # type: ignore[attr-defined]
        # Cascade read linked objects if necessary
        if cascade_read and not return_id:
            self._cascade_read(model_class, objs, return_copy)  # type: ignore[arg-type]
        return objs

    def read_one(
        self,
        model_class: type[Model],
        obj_id: Hashable,
        cascade_read: bool = False,
        return_id: bool = False,
        allow_duplicate_ids: bool = False,
        **kwargs: Any,
    ) -> Model | Hashable:
        return self.read_some(
            model_class,
            [obj_id],
            cascade_read=cascade_read,
            return_id=return_id,
            allow_duplicate_ids=allow_duplicate_ids,
            **kwargs,
        )[0]

    def read_some(
        self,
        model_class: type[Model],
        obj_ids: Iterable[Hashable],
        cascade_read: bool = False,
        return_id: bool = False,
        allow_duplicate_ids: bool = False,
        **kwargs: Any,
    ) -> list[Model] | list[Hashable]:
        return_copy = kwargs.get("return_copy", True)
        df = self._db[model_class]
        # Read some or one
        objs: list[Model] | list[Hashable]
        if return_id:
            objs = list(obj_ids)
            invalid_obj_ids = [x for x in objs if x not in df]
            if invalid_obj_ids:
                DictRepository._raise_invalid_ids(model_class, invalid_obj_ids)
            if not allow_duplicate_ids:
                DictRepository._verify_duplicate_ids(model_class, obj_ids)
        else:
            objs = [df.get(x) for x in obj_ids]  # type:ignore[assignment]
            # Verify input
            DictRepository._verify_valid_ids(model_class, obj_ids, objs)  # type: ignore[arg-type]
            if not allow_duplicate_ids:
                DictRepository._verify_duplicate_ids(model_class, obj_ids)

        # Make copy of objects for returning
        if not return_id and return_copy:
            objs = [x.model_copy() for x in objs if x]  # type: ignore[attr-defined]

        # Cascade read linked objects if necessary
        if cascade_read and not return_id:
            self._cascade_read(model_class, objs, return_copy)  # type: ignore[arg-type]
        return objs

    def upsert_some(
        self,
        user_id: Hashable,
        model_class: type[Model],
        objs: Model | Iterable[Model],
        raise_on_present: bool = False,
        raise_on_missing: bool = False,
        return_id: bool = False,
        **kwargs: Any,
    ) -> Hashable | list[Hashable] | Model | list[Model]:
        return_copy = kwargs.get("return_copy", True)
        df = self._db[model_class]
        get_id = self._get_id[model_class]
        is_iterable, objs = DictRepository._to_iterable(objs)
        obj_ids: list[Hashable] = [get_id(x) for x in objs]
        df_objs: list[Model | None] = [df.get(x) for x in obj_ids]
        # Verify input
        DictRepository._verify_duplicate_ids(model_class, obj_ids)
        invalid_type_ids = [get_id(x) for x in objs if not isinstance(x, model_class)]
        self._verify_upsert_objects(
            invalid_type_ids,
            model_class,
            obj_ids,
            df_objs,
            raise_on_present,
            raise_on_missing,
        )
        DictRepository._verify_duplicate_keys(
            get_id, self._keys_generators[model_class], model_class, objs, df.values()  # type: ignore[arg-type]
        )

        # Upsert objects
        value_field_names = self._value_field_names[model_class]
        links = [tuple([x[0], x[2], x[4]]) for x in self._links[model_class]]
        current_time = self._timestamp_factory()
        for i, obj, df_obj in zip(range(len(df_objs)), objs, df_objs):
            if df_obj:
                # Already existing -> update df_obj with obj data
                for field_name in value_field_names:
                    # Update value field
                    setattr(df_obj, field_name, getattr(obj, field_name))
                for link_field_name, relationship_field_name, linked_df in links:
                    # Verify and update link
                    linked_obj_id = getattr(obj, link_field_name)  # type: ignore[arg-type]
                    if not linked_obj_id:
                        setattr(df_obj, link_field_name, None)  # type: ignore[arg-type]
                        setattr(df_obj, relationship_field_name, None)  # type: ignore[arg-type]
                        continue
                    if linked_df is not None and linked_obj_id not in linked_df:
                        raise exc.InvalidIdsError(
                            (
                                f"Model {model_class}: obj {get_id(obj)} has invalid id"
                                f' in {link_field_name}: "{linked_obj_id}"'
                            ),
                            ids=[linked_obj_id],
                        )
                    setattr(df_obj, link_field_name, linked_obj_id)  # type: ignore[arg-type]
                    linked_obj = getattr(obj, relationship_field_name)  # type: ignore[arg-type]
                    if not linked_obj:
                        continue
                    get_link_id = self._get_id[linked_obj.__class__]
                    if get_link_id(linked_obj) != linked_obj_id:
                        raise exc.InvalidLinkIdsError(
                            (
                                f"Model {model_class}: obj {get_link_id(obj)} has different id "
                                f'in {link_field_name} ("{linked_obj_id}") versus '
                                f'{relationship_field_name} ("{get_link_id(linked_obj)}")'
                            ),
                            ids=[linked_obj_id, get_link_id(linked_obj)],
                        )
            else:
                # New -> insert copy of obj
                new_df_obj: Model = obj.model_copy()
                df[get_id(new_df_obj)] = new_df_obj
                df_objs[i] = new_df_obj
        if return_id:
            return obj_ids if is_iterable else obj_ids[0]
        if return_copy:
            return (
                [x.model_copy() for x in df_objs if x is not None]
                if is_iterable
                else df_objs[0].model_copy() if df_objs[0] is not None else None
            )
        return df_objs if is_iterable else df_objs[0]  # type: ignore[return-value]

    def delete_some(
        self,
        model_class: type[Model],
        obj_ids: Hashable | Iterable[Hashable],
        **kwargs: Any,
    ) -> Hashable | list[Hashable] | None:
        df = self._db[model_class]
        is_iterable, obj_ids = DictRepository._to_iterable(obj_ids)
        df_objs = [df.get(x) for x in obj_ids]
        back_links = self._back_links[model_class]

        # Verify input
        DictRepository._verify_valid_ids(model_class, obj_ids, df_objs)
        DictRepository._verify_duplicate_ids(model_class, obj_ids)

        # Verify existence of link (foreign key) constraint conflicts
        uq_obj_ids = set(obj_ids)
        for link_model_class, link_field_name in back_links:
            link_df = self._db[link_model_class]
            get_id = self._get_id[link_model_class]
            linked_obj_ids = [
                get_id(x)
                for x in link_df.values()
                if getattr(x, link_field_name) in uq_obj_ids
            ]
            if linked_obj_ids:
                linked_obj_ids_str = ", ".join([f'"{x}"' for x in linked_obj_ids])
                raise exc.LinkConstraintViolationError(
                    (
                        f"Model {model_class}: link constraint conflict in model "
                        f"{link_model_class}, id(s): {linked_obj_ids_str}"
                    ),
                    list(uq_obj_ids),
                    linked_obj_ids,
                )
        # Delete objects
        for obj_id in uq_obj_ids:
            df.pop(obj_id)
        return obj_ids if is_iterable else obj_ids[0]

    def delete_all(
        self, model_class: type[Model], filter: Filter | None, **kwargs: Any
    ) -> list[Hashable]:
        df = self._db[model_class]
        # Get any query filter
        query_filter: Filter | None = None
        obj_filter: Filter | None = kwargs.get("obj_filter")
        if filter and obj_filter:
            query_filter = CompositeFilter(
                filters=[filter, obj_filter], operator=LogicalOperator.AND  # type: ignore[list-item]
            )
        elif filter:
            query_filter = filter
        elif obj_filter:
            query_filter = obj_filter
        else:
            query_filter = None
        # Delete objects
        if query_filter:
            # Delete objects matching the query filter
            obj_ids = [
                x
                for x, y in zip(
                    df.keys(), query_filter.match_rows(df.values(), is_model=True)
                )
                if y
            ]
            for obj_id in obj_ids:
                self._db[model_class].pop(obj_id)
        else:
            # Delete all objects
            obj_ids = list(self._db[model_class].keys())
            self._db[model_class] = {}
        return obj_ids

    def exists_one(self, model_class: type[Model], obj_id: Hashable) -> bool:
        df = self._db[model_class]
        return obj_id in df

    def exists_some(
        self, model_class: type[Model], obj_ids: Iterable[Hashable]
    ) -> list[bool]:
        df = self._db[model_class]
        return [x in df for x in obj_ids]

    def _get_links(
        self, entity: Entity
    ) -> list[tuple[str, type[Model], str, int, dict[Hashable, Model] | None]]:
        # Return list[tuple[link_field_name, LinkModel, relationship_field_name, link_type_id, linked_df|None]]
        links = []
        for link_field_name in entity.get_link_field_names():
            (
                link_type_id,
                link_model_class,
                relationship_field_name,
            ) = entity.get_link_properties_by_field_name(link_field_name)
            links.append(
                (
                    link_field_name,
                    link_model_class,
                    relationship_field_name,
                    link_type_id,
                    self._db.get(link_model_class),  # type: ignore[arg-type]
                )
            )
        return links  # type: ignore[return-value]

    def _cascade_read(
        self,
        model_class: type[Model],
        objs: list[Model],
        return_copy: bool,
    ) -> None:
        for (
            link_field_name,
            link_model_class,
            relationship_field_name,
            _,
            linked_df,
        ) in self._links[model_class]:
            if linked_df is None:
                continue
            linked_obj_ids = [
                getattr(x, link_field_name) for x in objs if getattr(x, link_field_name)
            ]
            linked_objs = self.read_some(
                link_model_class,
                linked_obj_ids,
                cascade_read=False,
                allow_duplicate_ids=True,
                return_copy=return_copy,
            )
            for obj, linked_obj in zip(objs, linked_objs):
                setattr(obj, relationship_field_name, linked_obj)

    def _verify_upsert_objects(
        self,
        invalid_type_ids: list[Hashable],
        model_class: type[Model],
        obj_ids: list[Hashable],
        df_objs: list[Model | None],
        raise_on_present: bool,
        raise_on_missing: bool,
    ) -> None:
        if invalid_type_ids:
            invalid_type_ids_str = ", ".join([f'"{x}"' for x in invalid_type_ids])
            raise exc.InvalidModelIdsError(
                f"Model {model_class}: object(s) are of different type: {invalid_type_ids_str}",
                ids=invalid_type_ids,
            )
        get_id = self._get_id[model_class]
        if raise_on_present:
            present_ids = [get_id(x) for x in df_objs if x is not None]
            if present_ids:
                present_ids_str = ", ".join([f'"{x}"' for x in present_ids])
                raise exc.AlreadyExistingIdsError(
                    f"{model_class} object(s) already exist: {present_ids_str}",
                    ids=present_ids,
                )
        if raise_on_missing:
            missing_ids = [x for x, y in zip(obj_ids, df_objs) if y is None]
            if missing_ids:
                missing_ids_str = ", ".join([f"{x}" for x in missing_ids])
                raise exc.InvalidIdsError(
                    f"{model_class} object(s) do not exist: {missing_ids_str}",
                    ids=missing_ids,
                )

    @staticmethod
    def _verify_valid_ids(
        model_class: type[Model],
        obj_ids: Iterable[Hashable],
        objs: Iterable[Model | None],
    ) -> None:
        invalid_obj_ids = [x for x, y in zip(obj_ids, objs) if y is None]
        if invalid_obj_ids:
            DictRepository._raise_invalid_ids(model_class, invalid_obj_ids)

    @staticmethod
    def _raise_invalid_ids(
        model_class: type[Model], invalid_obj_ids: Iterable[Hashable]
    ) -> None:
        invalid_obj_ids_str = ", ".join(f'"{x}"' for x in invalid_obj_ids)
        raise exc.InvalidIdsError(
            f"Model {model_class}: invalid object id(s) provided: {invalid_obj_ids_str}",
            ids=invalid_obj_ids,
        )

    @staticmethod
    def _verify_duplicate_ids(
        model_class: type[Model], obj_ids: Iterable[Hashable]
    ) -> None:
        set_ = set()
        duplicate_ids = [x for x in obj_ids if x in set_ or set_.add(x)]  # type: ignore[func-returns-value]
        if duplicate_ids:
            DictRepository._raise_duplicate_ids(model_class, duplicate_ids)

    @staticmethod
    def _raise_duplicate_ids(
        model_class: type[Model], duplicate_ids: Iterable[Hashable]
    ) -> None:
        duplicate_ids_str = ", ".join([f'"{x}"' for x in duplicate_ids])
        raise exc.DuplicateIdsError(
            f"Model {model_class}: object ids are not unique: {duplicate_ids_str}",
            ids=duplicate_ids,
        )

    @staticmethod
    def _verify_duplicate_keys(
        get_id: Callable[[Model], Hashable],
        keys_generator: Callable,
        model_class: type[Model],
        objs: list[Model],
        df_objs: list[Model] | None,
    ) -> None:
        if not objs:
            # No objs -> no duplicates
            return
        keys = keys_generator(objs[0])
        if not keys:
            # No keys -> no duplicates
            return
        key_ids = list(keys.keys())

        def get_keys(obj: Any) -> Any:
            keys = keys_generator(obj)
            return tuple(keys[x] for x in key_ids)

        # Check for duplicate keys among objs
        obj_keys_list = [get_keys(x) for x in objs]
        n_keys = len(keys)
        duplicate_objs = []
        for i in range(n_keys):
            curr_obj_keys_list = [x[i] for x in obj_keys_list]
            curr_obj_keys = set(curr_obj_keys_list)
            if len(curr_obj_keys) < len(curr_obj_keys_list):
                seen = set()
                uq_obj_keys = set(
                    x for x in curr_obj_keys_list if x not in seen and not seen.add(x)  # type: ignore[func-returns-value]
                )
                duplicate_obj_keys = curr_obj_keys - uq_obj_keys
                duplicate_objs += [
                    x
                    for x, y in zip(objs, curr_obj_keys_list)
                    if y in duplicate_obj_keys
                ]
        if duplicate_objs:
            raise exc.UniqueConstraintViolationError(
                f"Model {model_class}: object keys are not unique",
                duplicate_key_ids=list(set([get_id(x) for x in duplicate_objs])),
            )
        # Check for duplicate keys between objs and df_objs, excluding those df_objs
        #  that have the same id as an obj
        if not df_objs:
            return
        obj_ids = set([get_id(x) for x in objs])
        df_obj_keys = [get_keys(x) for x in df_objs if get_id(x) not in obj_ids]
        duplicate_objs = []
        for i in range(n_keys):
            curr_df_obj_keys = set(x[i] for x in df_obj_keys)
            curr_obj_keys_list = [x[i] for x in obj_keys_list]
            curr_obj_keys = set(curr_obj_keys_list)
            duplicate_obj_keys = curr_obj_keys & curr_df_obj_keys
            if duplicate_obj_keys:
                duplicate_objs += [
                    x
                    for x, y in zip(objs, curr_obj_keys_list)
                    if y in duplicate_obj_keys
                ]
        if duplicate_objs:
            raise exc.UniqueConstraintViolationError(
                f"Model {model_class}: object keys are not unique",
                duplicate_key_ids=list(set([get_id(x) for x in duplicate_objs])),
            )

    @staticmethod
    def _to_iterable(
        obj: Any | Iterable[Any],
    ) -> tuple[bool, list[Any] | set[Any] | frozenset[Any]]:
        # TODO: take model_class as argument, and derive id field class from it to use for check
        if isinstance(obj, (list, set, frozenset)):
            return True, obj
        if issubclass(type(obj), Model):
            return False, [obj]
        if isinstance(obj, UUID):
            return False, [obj]
        try:
            list_obj = list(obj)
            if not list_obj:
                return True, list_obj
            if issubclass(type(list_obj[0]), Model):
                return False, list_obj
            return False, [obj]
        except:
            return False, [obj]



from gen_epix.commondb.domain.command import COMMANDS_BY_SERVICE_TYPE
from gen_epix.commondb.domain.model import (
    SORTED_MODELS_BY_SERVICE_TYPE,
    SORTED_SERVICE_TYPES,
)
from gen_epix.commondb.util import register_domain_entities
from gen_epix.fastapp import Domain

DOMAIN = Domain("commondb")

register_domain_entities(
    DOMAIN,
    SORTED_SERVICE_TYPES,
    SORTED_MODELS_BY_SERVICE_TYPE,  # type: ignore[arg-type]
    COMMANDS_BY_SERVICE_TYPE,  # type: ignore[arg-type]
    set_schema_to_service_type=True,
)



import abc

from gen_epix.commondb.domain import model
from gen_epix.fastapp import BaseRepository, BaseUnitOfWork


class BaseOrganizationRepository(BaseRepository):
    def __init__(
        self,
        user_class: type[model.User] = model.User,
        user_invitation_class: type[model.UserInvitation] = model.UserInvitation,
    ):
        super().__init__()
        self.user_class = user_class
        self.user_invitation_class = user_invitation_class

    @abc.abstractmethod
    def is_existing_user_by_key(
        self, uow: BaseUnitOfWork, user_key: str | None
    ) -> bool:
        raise NotImplementedError()

    @abc.abstractmethod
    def retrieve_user_by_key(self, uow: BaseUnitOfWork, user_key: str) -> model.User:
        raise NotImplementedError()



# pylint: disable=useless-import-alias
from gen_epix.casedb.domain.model.case.complete_case_type import (
    CompleteCaseType as CompleteCaseType,
)
from gen_epix.casedb.domain.model.case.non_persistable import (
    BaseCaseRights as BaseCaseRights,
)
from gen_epix.casedb.domain.model.case.non_persistable import (
    CaseDataIssue as CaseDataIssue,
)
from gen_epix.casedb.domain.model.case.non_persistable import CaseQuery as CaseQuery
from gen_epix.casedb.domain.model.case.non_persistable import CaseReadSet as CaseReadSet
from gen_epix.casedb.domain.model.case.non_persistable import CaseRights as CaseRights
from gen_epix.casedb.domain.model.case.non_persistable import CaseSeq as CaseSeq
from gen_epix.casedb.domain.model.case.non_persistable import (
    CaseSetQuery as CaseSetQuery,
)
from gen_epix.casedb.domain.model.case.non_persistable import (
    CaseSetRights as CaseSetRights,
)
from gen_epix.casedb.domain.model.case.non_persistable import CaseSetStat as CaseSetStat
from gen_epix.casedb.domain.model.case.non_persistable import CaseTypeDim as CaseTypeDim
from gen_epix.casedb.domain.model.case.non_persistable import (
    CaseTypeStat as CaseTypeStat,
)
from gen_epix.casedb.domain.model.case.non_persistable import (
    CaseValidationReport as CaseValidationReport,
)
from gen_epix.casedb.domain.model.case.non_persistable import (
    ValidatedCase as ValidatedCase,
)
from gen_epix.casedb.domain.model.case.persistable import Case as Case
from gen_epix.casedb.domain.model.case.persistable import (
    CaseDataCollectionLink as CaseDataCollectionLink,
)
from gen_epix.casedb.domain.model.case.persistable import CaseSet as CaseSet
from gen_epix.casedb.domain.model.case.persistable import (
    CaseSetCategory as CaseSetCategory,
)
from gen_epix.casedb.domain.model.case.persistable import (
    CaseSetDataCollectionLink as CaseSetDataCollectionLink,
)
from gen_epix.casedb.domain.model.case.persistable import CaseSetMember as CaseSetMember
from gen_epix.casedb.domain.model.case.persistable import CaseSetStatus as CaseSetStatus
from gen_epix.casedb.domain.model.case.persistable import CaseType as CaseType
from gen_epix.casedb.domain.model.case.persistable import CaseTypeCol as CaseTypeCol
from gen_epix.casedb.domain.model.case.persistable import (
    CaseTypeColSet as CaseTypeColSet,
)
from gen_epix.casedb.domain.model.case.persistable import (
    CaseTypeColSetMember as CaseTypeColSetMember,
)
from gen_epix.casedb.domain.model.case.persistable import CaseTypeSet as CaseTypeSet
from gen_epix.casedb.domain.model.case.persistable import (
    CaseTypeSetCategory as CaseTypeSetCategory,
)
from gen_epix.casedb.domain.model.case.persistable import (
    CaseTypeSetMember as CaseTypeSetMember,
)
from gen_epix.casedb.domain.model.case.persistable import Col as Col
from gen_epix.casedb.domain.model.case.persistable import Dim as Dim
from gen_epix.casedb.domain.model.case.persistable import (
    GeneticDistanceProtocol as GeneticDistanceProtocol,
)
from gen_epix.casedb.domain.model.case.persistable import TreeAlgorithm as TreeAlgorithm
from gen_epix.casedb.domain.model.case.persistable import (
    TreeAlgorithmClass as TreeAlgorithmClass,
)



from gen_epix.fastapp import BaseService
from gen_epix.omopdb.domain.enum import ServiceType
from gen_epix.omopdb.domain.repository.omop import BaseOmopRepository


class BaseOmopService(BaseService):
    SERVICE_TYPE = ServiceType.OMOP

    # Property overridden to provide narrower return value to support linter
    @property  # type: ignore
    def repository(self) -> BaseOmopRepository:  # type: ignore
        return super().repository  # type: ignore

    def register_handlers(self) -> None:
        self.register_default_crud_handlers()



from gen_epix.commondb.domain.repository.system import BaseSystemRepository
from gen_epix.fastapp.repositories import SARepository


class SystemSARepository(SARepository, BaseSystemRepository):
    pass



import abc

from gen_epix.commondb.domain import command, model
from gen_epix.commondb.domain.enum import ServiceType
from gen_epix.commondb.domain.repository.system import BaseSystemRepository
from gen_epix.fastapp import BaseService


class BaseSystemService(BaseService):
    SERVICE_TYPE = ServiceType.SYSTEM

    # Property overridden to provide narrower return value to support linter
    @property  # type: ignore
    def repository(self) -> BaseSystemRepository:  # type: ignore
        return super().repository  # type: ignore

    def register_handlers(self) -> None:
        f = self.app.register_handler
        self.register_default_crud_handlers()
        f(command.RetrieveOutagesCommand, self.retrieve_outages)
        f(command.RetrieveLicensesCommand, self.retrieve_licenses)

    @abc.abstractmethod
    def register_policies(self) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def retrieve_outages(
        self, cmd: command.RetrieveOutagesCommand
    ) -> list[model.Outage]:
        raise NotImplementedError

    @abc.abstractmethod
    def retrieve_licenses(
        self, cmd: command.RetrieveLicensesCommand
    ) -> list[model.PackageMetadata]:
        raise NotImplementedError



from enum import Enum


class TimeUnit(Enum):
    YEAR = "YEAR"
    QUARTER = "QUARTER"
    MONTH = "MONTH"
    WEEK = "WEEK"
    DAY = "DAY"


class TimeUnitTransformStrategy(Enum):
    EXACT_ONLY = "EXACT_ONLY"
    LARGEST_OVERLAP = "LARGEST_OVERLAP"


class NoMatchStrategy(Enum):
    RAISE = "RAISE"
    SET_NONE = "SET_NONE"


class TransformType(Enum):
    BASE = "BASE"


class TransformResultType(Enum):
    """Enum for different types of transformation results."""

    SUCCESS = "SUCCESS"
    ERROR = "ERROR"
    SKIPPED = "SKIPPED"



# pylint: disable=useless-import-alias
from gen_epix.casedb.domain.model.abac.policy import (
    OrganizationAccessCasePolicy as OrganizationAccessCasePolicy,
)
from gen_epix.casedb.domain.model.abac.policy import (
    OrganizationShareCasePolicy as OrganizationShareCasePolicy,
)
from gen_epix.casedb.domain.model.abac.policy import (
    UserAccessCasePolicy as UserAccessCasePolicy,
)
from gen_epix.casedb.domain.model.abac.policy import (
    UserShareCasePolicy as UserShareCasePolicy,
)
from gen_epix.casedb.domain.model.abac.rights import CaseAbac as CaseAbac
from gen_epix.casedb.domain.model.abac.rights import (
    CaseTypeAccessAbac as CaseTypeAccessAbac,
)
from gen_epix.casedb.domain.model.abac.rights import (
    CaseTypeShareAbac as CaseTypeShareAbac,
)
from gen_epix.commondb.domain.model import (
    OrganizationAdminPolicy as OrganizationAdminPolicy,
)



# pylint: disable=useless-import-alias
from gen_epix.commondb.services import AuthService as AuthService
from gen_epix.commondb.services import OrganizationService as OrganizationService
from gen_epix.commondb.services import RbacService as RbacService
from gen_epix.commondb.services import SystemService as SystemService
from gen_epix.commondb.services import UserManager as UserManager
from gen_epix.omopdb.services.abac import AbacService as AbacService
from gen_epix.omopdb.services.omop import OmopService as OmopService
from gen_epix.omopdb.services.organization import (
    OrganizationService as OrganizationService,
)
from gen_epix.omopdb.services.rbac import RbacService as RbacService
